/**
 * AnimationPlayer.java (NoFrills3D)
 * Created by Joshua Newnham on Apr 29, 2013.
 * Copyright (c) 2013 We Make Play All rights reserved.
 */
package com.wmp.nofrills3d.animations;

import java.util.List;

import com.wmp.nofrills3d.geometry.Bone;
import com.wmp.nofrills3d.geometry.SkinnedMesh;
import com.wmp.nofrills3d.math.Matrix4;

public class AnimationPlayer {		

	protected SkinnedMesh skinnedMesh; 
	
	protected Matrix4[] boneTransforms;	
    
	protected float currentTime = 0.0f; 
	
	protected int currentKeyframe = 0;
	
	protected AnimationClipSequence currentAnimation;
	
	protected boolean isBlending = false;

	protected float blendTime = 0.0f;

	protected float blendElapsedTime = 0.0f;

	protected AnimationClipSequence blendAnimation = null;
	
	protected IAnimationDelegate delegate; 
	
	public boolean playing = false; 
	
	public AnimationPlayer(SkinnedMesh skinnedMesh, Matrix4[] boneTransforms){
		this.skinnedMesh = skinnedMesh; 
		this.boneTransforms = boneTransforms; 
	}
	
	public AnimationPlayer(SkinnedMesh skinnedMesh, Matrix4[] boneTransforms, IAnimationDelegate delegate){
		this(skinnedMesh, boneTransforms);
		this.delegate = delegate; 
	}
	
	public void startAnimation(AnimationClipSequence animationClip){
		currentAnimation = animationClip; 
		currentTime = 0.0f; 
		currentKeyframe = 0; 
		playing = true; 
		
		resetToFirstFrame(); 
	}
	
	public void crossFadeAnimation(AnimationClipSequence animationClip, float blendTime){
		if( currentAnimation == null ){
			startAnimation(animationClip);
		} else{
			blendAnimation = animationClip; 
			
			this.blendTime = blendTime;			
			this.isBlending = true; 
			this.blendElapsedTime = 0.0f;		
		}
	}
	
	public void resetToBindPose(){
		skinnedMesh.copyBindPose(boneTransforms);
	}
	
	public void resetToFirstFrame(){
		if( currentAnimation == null ){
			return; 
		}
		
		List<Keyframe> keyframes = currentAnimation.animationClip.keyframes;
		float startTime = currentAnimation.startTime;
		
		for( int i=0; i<keyframes.size(); i++ ){
			Keyframe keyframe = keyframes.get(i);
			
			if( keyframe.time > startTime ){
				break; 
			}
			
			if( skinnedMesh.rootBone.findBoneAtIndex(keyframe.bone.index).isLocked() ){
        		continue; 
        	}
			
			// Use this keyframe.
            boneTransforms[keyframe.bone.index].set(keyframe.transform);
		}			
	}
	
	public float getTimeRemaining(){
		if( currentAnimation != null ){
			return (currentAnimation.endTime - currentTime)/currentAnimation.speed;
		} 
		return 0.0f; 
	}
	
	public void update(float elapsedTime){
		if( currentAnimation == null || !playing ){
			return; 
		}
		
		float dTime = (elapsedTime * currentAnimation.speed); 
		
		// advance time 		
		currentTime += dTime;
		
		if (isBlending)
        {                
            blendElapsedTime += dTime;
        }
		
		// if current time has past the end of the animation 
        while (currentTime >= currentAnimation.getDuration())
        {
            // if we're looping then reduce the time until we are back in the animation start time frame 
            if (currentAnimation.looping && !isBlending)
            {
                currentTime -= currentAnimation.getDuration();
                currentKeyframe = 0;
                
                resetToFirstFrame();                               
            }
            else
            { // other wise clamp to the end of the animation 
                if( !isBlending ){
                    playing = false; 
                }

                currentTime = currentAnimation.endTime; 
                break;
            }
        }
        
        float[] wrkTransform = new float[16];
        
        // read the keyframe matricies 
        List<Keyframe> keyframes = currentAnimation.animationClip.keyframes;
        int[][] boneKeyframes = currentAnimation.animationClip.boneKeyframes;
        
        float frameTime = currentTime + currentAnimation.startTime;        
        
        long currentSystemTime = System.currentTimeMillis(); 
        
        for( int boneIndex=0; boneIndex < boneKeyframes.length; boneIndex++ ){
        	
        	Bone currentBone = skinnedMesh.rootBone.findBoneAtIndex(boneIndex);  
        	
        	if( currentBone.isLocked() ){
        		continue; 
        	}
        	
        	float lockBlendTime = 1.0f; 
        	
        	// test if blending is required (blending is done once a lock has been removed) 
        	long diffLockTime = currentSystemTime - currentBone.getLockChangedTimestamp();
        	if( diffLockTime < currentBone.lockBlendTime ){  
        		lockBlendTime = (float)diffLockTime/(float)currentBone.lockBlendTime;
        	}
        	
        	int[] keyframeIndex = boneKeyframes[boneIndex];
        	
        	for( int i=0; i<keyframeIndex.length; i++ ){
        		Keyframe keyframe = keyframes.get(keyframeIndex[i]);
        		        		        		
        		if( keyframe.time > frameTime ){        			        			
        			Keyframe prevKeyframe = keyframes.get(keyframeIndex[(i-1 >= 0 ? i-1 : keyframeIndex.length - 1)]);
        			
        			prevKeyframe.copyTransform(wrkTransform); 
        			if(  frameTime > prevKeyframe.time ){
        				
        				// get next frame
        				if( keyframe.time <= currentAnimation.endTime ){
        					float t = ((frameTime - prevKeyframe.time)/(keyframe.time - prevKeyframe.time));
        					interpolate(wrkTransform, keyframe.transform, t, wrkTransform);
        				} else{
        					if( currentAnimation.looping ){

        					}
        				}
        			}
        			
        			if( lockBlendTime < 0.9 ){
        				interpolate(boneTransforms[prevKeyframe.bone.index].matrix, wrkTransform, lockBlendTime, boneTransforms[prevKeyframe.bone.index].matrix);        				
        			} else{
        				boneTransforms[prevKeyframe.bone.index].set(wrkTransform);
        			}
        			break; 
        		}
        	}        	
        }
        
        if (isBlending)
        {
            // blend current animation with blend aniamtion 
            // find out how much blend we're doing 
            float r = blendTime - blendElapsedTime;
            float blend = 1.0f; // amount we're blending by 
            if (r > 0)
            {
                blend = 1 - (r / blendTime); 
            }

            // blend together the bone transform          
            List<Keyframe> blendKeyframes = blendAnimation.animationClip.keyframes;
            
            for( int boneIndex=0; boneIndex < boneKeyframes.length; boneIndex++ ){
            	
            	if( skinnedMesh.rootBone.findBoneAtIndex(boneIndex).isLocked() ){
            		continue; 
            	}
            	
            	int[] keyframeIndex = boneKeyframes[boneIndex];
            	
            	for( int i=0; i<keyframeIndex.length; i++ ){
            		Keyframe keyframe = blendKeyframes.get(keyframeIndex[i]);
            		        		        		
            		if( keyframe.time > blendAnimation.startTime ){        			        			
            			Keyframe prevKeyframe = blendKeyframes.get(keyframeIndex[(i-1 >= 0 ? i-1 : keyframeIndex.length - 1)]);
            			
            			prevKeyframe.copyTransform(wrkTransform);             			
            			
            			interpolate(wrkTransform, keyframe.transform, blend, wrkTransform);
            			
            			boneTransforms[prevKeyframe.bone.index].set(wrkTransform);
            			break; 
            		}
            	}            
            }

            // are we finished blending?
            if (r < 0.0f)
            {
                // transfer blend to current animation clip 
                currentAnimation = blendAnimation; 

                // reset variables 
                currentTime = 0.0f;
                currentKeyframe = 0;
                playing = true;                

                // clear blend variables 
                isBlending = false;
                blendAnimation = null;  
                
                isBlending = false; 
            }

        }  
        
        if( !playing && delegate != null ){
        	delegate.onAnimationFinished(currentAnimation);
        }
	}
	
	protected void interpolate(float[] src, float[] dst, float t, float[] res){
		for( int i=0; i<src.length; i++ ){
			res[i] = src[i] + ((dst[i]-src[i]) * t); 
		}
	}
}
