/**
 * Camera.java (ColourExplosion)
 * Created by Joshua Newnham on Apr 9, 2013.
 * Copyright (c) 2013 We Make Play All rights reserved.
 */
package com.wmp.nofrills3d.entities;

import static android.opengl.Matrix.setLookAtM;
import android.graphics.Point;
import android.opengl.Matrix;

import com.wmp.nofrills3d.math.Matrix4;
import com.wmp.nofrills3d.math.Ray;
import com.wmp.nofrills3d.math.Vector3;

public abstract class Camera extends Entity {
	
	public int viewWidth = 0; 
	public int viewHeight = 0; 
	
	public Vector3 up = new Vector3(0, 1, 0);
	
	public Matrix4 projectionMatrix = new Matrix4(); 
	
	public Matrix4 viewProjectionMatrix = Matrix4.createIdentity(); 
	
	public Matrix4 invertedViewProjectionMatrix = new Matrix4(); 
	
	public Camera(){
		
	}
	
	public void lookAt(Vector3 target){
		setLookAtM(modelTransform.matrix, 0, 
				position.x, position.y, position.z, 
				target.x, target.y, target.z, 
				up.x, up.y, up.z);
		
		// invert matrix 
		invModelTransform.reset();
		Matrix4.invert(modelTransform, invModelTransform);
		
		modelTransform.set(invModelTransform.matrix);
		
		Matrix4.decompose(modelTransform, position, rotation, scale);
	}	
	
	public abstract boolean init(); 
	
	@Override
	public void refreshModelMatrix(){
		//super.refreshModelMatrix();
		
		modelTransform.reset(); 

		Matrix.translateM(modelTransform.matrix, 0, position.x, position.y, position.z);
		Matrix.rotateM(modelTransform.matrix, 0, rotation.x, 1.0f, 0.0f, 0.0f);
		Matrix.rotateM(modelTransform.matrix, 0, rotation.y, 0.0f, 1.0f, 0.0f);
		Matrix.rotateM(modelTransform.matrix, 0, rotation.z, 0.0f, 0.0f, 1.0f);		
		
		invModelTransform.reset();
		Matrix4.invert(modelTransform, invModelTransform);
		
		modelTransform.set(invModelTransform.matrix);
		
		// Update the viewProjection matrix, and create an inverted matrix for touch picking.
        Matrix4.multiply(projectionMatrix, modelTransform, viewProjectionMatrix);
        Matrix4.invert(viewProjectionMatrix, invertedViewProjectionMatrix);
	}
	
	public Ray screenPositionToRay(float x, float y){
		// Convert touch coordinates into normalized device coordinates, keeping in mind that Android's Y coordinates are inverted.
        final float normalizedX = normalizeX(x);
        final float normalizedY = normalizeY(y);
            
        // convert normalized 2D point to a Ray
        
        // We'll convert these normalized device coordinates into world-space
		// coordinates. We'll pick a point on the near and far planes, and draw a
		// line between them. To do this transform, we need to first multiply by
		// the inverse matrix, and then we need to undo the perspective divide.
		final float[] nearPointNdc = { normalizedX, normalizedY, -1, 1 };
		final float[] farPointNdc = { normalizedX, normalizedY, 1, 1 };		

		final float[] nearPointWorld = new float[4];
		final float[] farPointWorld = new float[4];

		Matrix.multiplyMV(nearPointWorld, 0, invertedViewProjectionMatrix.matrix, 0, nearPointNdc, 0);
		Matrix.multiplyMV(farPointWorld, 0, invertedViewProjectionMatrix.matrix, 0, farPointNdc, 0);

		// Why are we dividing by W? We multiplied our vector by an inverse
		// matrix, so the W value that we end up is actually the *inverse* of
		// what the projection matrix would create. By dividing all 3 components
		// by W, we effectively undo the hardware perspective divide.
		divideByW(nearPointWorld);
		divideByW(farPointWorld);

		// We don't care about the W value anymore, because our points are now
		// in world coordinates.
		Vector3 nearPointRay = new Vector3(nearPointWorld[0], nearPointWorld[1], nearPointWorld[2]);
		Vector3 farPointRay = new Vector3(farPointWorld[0], farPointWorld[1], farPointWorld[2]);

		return new Ray(nearPointRay, Vector3.subtract(nearPointRay, farPointRay));
	}
	
	protected float normalizeX(float x){
		return (x / (float) viewWidth) * 2 - 1;
	}
	
	protected float normalizeY(float y){
		return -((y / (float) viewHeight) * 2 - 1);
	}	

	protected void divideByW(float[] vector) {
		vector[0] /= vector[3];
		vector[1] /= vector[3];
		vector[2] /= vector[3];
	}
	
}
