/**
 * NoFrills3D
 * com.wmp.nofrills3d.entities
 * 
 * @author Joshua Newnham
 * @copyright We Make Play Ltd. 2013 
 * 
 */
package com.wmp.nofrills3d.entities;

import java.util.ArrayList;
import java.util.List;

import com.wmp.nofrills3d.math.Matrix4;
import com.wmp.nofrills3d.physics.ContactCollection;
import com.wmp.nofrills3d.physics.ICollider;
import com.wmp.nofrills3d.scene.IScene;

/**
 * 
 */
public class EntityGroup extends Entity {

	public List<Entity> children = new ArrayList<Entity>();  
	
	public EntityGroup(){
		super(); 
	}
	
	public void addChild(Entity newChild){
		this.children.add(newChild);
		
		if( scene != null ){
			for( Entity child : children ){
				if( !child.isLoaded() ){
					child.load(scene);
				}
			}
		}
	}
	
	public void removeChild(Entity child){
		this.children.remove(child);
	}
	
	@Override
	public void load(IScene scene){
		super.load(scene);
				
		for( Entity child : children ){
			if( !child.isLoaded() ){
				child.load(scene);
			}
		}
	}		
	
	@Override
	public void unload(boolean fullDispose){
		for( Entity child : children ){
			child.unload(fullDispose); 
		}
	}	
	
	@Override
	public boolean isLoaded(){
		for( Entity child : children ){
			if( !child.isLoaded() ){
				return false; 
			}
		}
		
		return super.isLoaded(); 
	}
	
	@Override
	public void update(float elapsedTime){
		super.update(elapsedTime);
		
		for( Entity child : children ){
			child.update(elapsedTime);
			
			Matrix4.multiply(this.modelTransform, child.modelTransform, child.modelTransform); 
		}				
	}
	
	@Override
	public void postUpdate(float elapsedTime){
		super.postUpdate(elapsedTime);
		
		colliderCollection.update(elapsedTime);
		
		for( Entity child : children ){
			child.postUpdate(elapsedTime);
		}
	}
	
	@Override
	public void render(Matrix4 projectionMatrix, Matrix4 viewTransform, int pass){
		for( Entity child : children ){
			child.render(projectionMatrix, viewTransform, pass);
		}
	}	
	
	@Override
	public void onHandleCollisions(ContactCollection contacts){
		for( Entity child : children ){
			child.onHandleCollisions(contacts);
		}
	}
	
	@Override
	public ICollider getCollidableAtIndex(int index) {
		ICollider collider = colliderCollection.getCollidableAtIndex(index);
		
		if( collider != null ){
			return collider; 
		}
		
		if( collider == null ){			
			for( Entity child : children ){
				collider = child.getCollidableAtIndex(index);
				
				if( collider != null ){
					return collider; 
				}
			}
		}
		
		return null; 
	}

	@Override
	public ICollider getCollidableWithTag(String tag) {
		ICollider collider = colliderCollection.getCollidableWithTag(tag);				
		
		if( collider != null ){
			return collider; 
		}
		
		if( collider == null ){			
			for( Entity child : children ){
				collider = child.getCollidableWithTag(tag);
				
				if( collider != null ){
					return collider; 
				}
			}
		}
		
		return null; 
	}
	
	@Override
	public boolean isCollidable(){
		int localColliders = colliderCollection.size();
		
		if( localColliders > 0 ){
			return true; 
		}
		
		for( Entity child : children ){
			if( child.isCollidable() ){
				return true; 
			}
		}
		
		return false; 
	}
	
	@Override
	public int collidesWith(ICollider other, ContactCollection contacts) {
		int count = colliderCollection.collidesWith(other, contacts);
		
		for( Entity child : children ){
			count += child.collidesWith(other, contacts);
		}
		
		return count; 
	}	
	
	public int collidesWith(Entity other, ContactCollection contacts) {
		int sum = 0; 
		
		for( ICollider collider : colliderCollection ){
			for( ICollider otherCollider : other.colliderCollection ){
				sum += collider.collidesWith(otherCollider, contacts);
			}
		}
		
		for( Entity child : children ){
			sum += child.collidesWith(other, contacts);
		}
		
		return sum; 
	}
	
}
