/**
 * SkinnedModel.java (NoFrills3D)
 * Created by Joshua Newnham on Apr 12, 2013.
 * Copyright (c) 2013 We Make Play All rights reserved.
 */
package com.wmp.nofrills3d.entities;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import com.wmp.nofrills3d.animations.AnimationClip;
import com.wmp.nofrills3d.animations.AnimationClipSequence;
import com.wmp.nofrills3d.animations.AnimationPlayer;
import com.wmp.nofrills3d.animations.IAnimationDelegate;
import com.wmp.nofrills3d.geometry.Bone;
import com.wmp.nofrills3d.geometry.IMesh;
import com.wmp.nofrills3d.geometry.SkinnedMesh;
import com.wmp.nofrills3d.ik.IKConstraint;
import com.wmp.nofrills3d.material.Shader;
import com.wmp.nofrills3d.material.Texture;
import com.wmp.nofrills3d.material.Material;
import com.wmp.nofrills3d.material.SkinnedMaterial;
import com.wmp.nofrills3d.math.Matrix4;
import com.wmp.nofrills3d.physics.BoneCollider;
import com.wmp.nofrills3d.physics.BoxCollider;
import com.wmp.nofrills3d.physics.ICollider;
import com.wmp.nofrills3d.utils.Logger;

public class SkinnedModel extends Model implements IAnimationDelegate {
	
	public static final String TAG = "SkinnedModel";
	
	public Matrix4[] boneTransforms; 
	
	public Matrix4[] worldTransforms; 
	
	public Matrix4[] skinTransforms;
	
	protected float[] skinTransformMatrices;
	
	protected int boneCount = 0; 
	
	protected Hashtable<String, AnimationClipSequence> animationSequences;  
	
	protected AnimationPlayer animationPlayer; 
	
	protected List<IKConstraint> ikChains = new Vector<IKConstraint>(); 
	
	public SkinnedModel(){
		super(); 
	}
	
	public SkinnedModel(SkinnedMesh mesh, SkinnedMaterial material){
		initModel(mesh, material, null);		
	}
	
	public SkinnedModel(SkinnedMesh mesh, SkinnedMaterial material, Texture texture){
		initModel(mesh, material, texture);
	}	
	
	@Override
	protected void initModel(IMesh mesh, Material material, Texture texture){
		super.initModel(mesh, material, texture);
		
		initTransforms(); 
		
		SkinnedMesh skinnedMesh = (SkinnedMesh)mesh;
		
		if( initAnimationPlayer(skinnedMesh) ){
			initAnimationClipSequences(skinnedMesh);
		}	
		
		if( skinnedMesh.getBoundingBox() != null ){
			skinnedMesh.getBoundingBox().setTransform(this);
		}
	}
	
	protected void initTransforms(){
		SkinnedMesh skinnedMesh = (SkinnedMesh)mesh;
		
		boneCount = skinnedMesh.bindPose.length; 
		
		boneTransforms = new Matrix4[boneCount]; 
		for( int i=0; i<boneCount; i++ ){
			boneTransforms[i] = skinnedMesh.bindPose[i].clone(); 
		}
		
		worldTransforms = new Matrix4[boneCount];
		for( int i=0; i<boneCount; i++ ){
			worldTransforms[i] = Matrix4.createIdentity();  
		}
		
		skinTransforms = new Matrix4[boneCount];
		for( int i=0; i<boneCount; i++ ){
			skinTransforms[i] = Matrix4.createIdentity();  
		}
		
		skinTransformMatrices = new float[boneCount * 16];
		
		//logBoneHierarchy(skinnedMesh);
	}
	
	protected boolean initAnimationPlayer(SkinnedMesh skinnedMesh){				
		if( skinnedMesh.animationClips == null || skinnedMesh.animationClips.size() == 0 ){
			return false; 
		}
		
		animationSequences = new Hashtable<String, AnimationClipSequence>(); 
		animationPlayer = new AnimationPlayer(skinnedMesh, boneTransforms, this);	
		
		return true; 
	}
	
	
	protected void initAnimationClipSequences(SkinnedMesh skinnedMesh){		
		
		AnimationClip animationClip = skinnedMesh.animationClips.get(0); 
		animationClip.init(skinnedMesh.rootBone); 				
		
		AnimationClipSequence defaultAnimation = new AnimationClipSequence(animationClip);		
		animationSequences.put("default", defaultAnimation);
		animationPlayer.startAnimation(defaultAnimation);
	}
	
	public Bone findBoneAtIndex(int index){
		return ((SkinnedMesh)mesh).rootBone.findBoneAtIndex(index);
	}
	
	public Bone findBoneWithId(String id){
		return ((SkinnedMesh)mesh).rootBone.findBoneWithId(id);
	}
	
	@Override
	public void update(float elapsedTime){
		super.update(elapsedTime);
		
		updateBoneTransforms(elapsedTime);								
		
		updateWorldTransforms(Matrix4.createIdentity());
		
		updateIKChains(elapsedTime); 
		
		updateSkinTransforms();
		
		copySkinTransformToSkinArray();
	}	
	
	@Override
	public void render(Matrix4 projectionMatrix, Matrix4 viewTransform, int pass){
		if( !canRender() ) return; 				
		
		// bind material 
		material.modelTransform = this.modelTransform; 
		material.viewTransform = viewTransform; 
		material.projectionTransform = projectionMatrix; 
		material.texture = texture;
		material.pass = pass; 
		((SkinnedMaterial)material).setBoneTransforms(skinTransformMatrices);
		
		material.bind(); 
		mesh.bind(material); 
		
		mesh.draw(); 
		
		mesh.unbind(material); 
		material.unbind(); 		
	}
	
	protected void updateBoneTransforms(float elapsedTime){
		if( animationPlayer != null && animationPlayer.playing ){
			animationPlayer.update(elapsedTime);
		}
	}
	
	protected void updateIKChains(float elapsedTime){
		for( IKConstraint ik : ikChains ){
			ik.update(elapsedTime);
		}
	}
	
	protected void updateWorldTransforms(Matrix4 rootTransform){ 
		Matrix4.multiply(rootTransform, boneTransforms[0], worldTransforms[0]);
		
		SkinnedMesh skinnedMesh = (SkinnedMesh)mesh; 
		
		// for each child bone 
		for( int bone=1; bone<boneCount; bone++ ){			
			int parentBoneIndex = skinnedMesh.boneHierarchy[bone];
			Matrix4.multiply(worldTransforms[parentBoneIndex], boneTransforms[bone], worldTransforms[bone]);
		}
	}
	
	protected void updateSkinTransforms(){
		SkinnedMesh skinnedMesh = (SkinnedMesh)mesh; 
		
		for( int bone=0; bone<boneCount; bone++ ){			
			Matrix4 invBindPose = skinnedMesh.inverseBindPose[bone];
			Matrix4.multiply(worldTransforms[bone], invBindPose, skinTransforms[bone]);
		}
	}
	
	protected void copySkinTransformToSkinArray(){
		// pack our skin transforms into a single array
		int arrayIndex = 0; 
		for( int i=0; i<boneCount; i++ ){
			Matrix4 skinMatrix =  skinTransforms[i];
			for( int j=0; j<16; j++ ){
				skinTransformMatrices[arrayIndex] = skinMatrix.matrix[j];
				arrayIndex++; 
			}			
		}
	}
	
	protected void logBoneHierarchy(SkinnedMesh mesh){
		StringBuffer sb = new StringBuffer(); 
		printBoneHierarchy(mesh, mesh.rootBone, sb, 1);
		Logger.Log(TAG, sb.toString());
	}
	
	protected void printBoneHierarchy(SkinnedMesh mesh, Bone bone, StringBuffer sb, int level){
		sb.append(String.format("\n%" + level + "s", ""));		
		sb.append("("); sb.append(mesh.boneHierarchy[bone.index]); sb.append(" "); sb.append(bone.index); sb.append(")"); sb.append(" "); sb.append(bone.id);
		
		if( bone.children.size() > 0 ){
			sb.append(" -> ");
		}
		
		for(Bone child : bone.children){
			printBoneHierarchy(mesh, child, sb, level+1);
		}	
	}

	@Override
	public void onAnimationFinished(AnimationClipSequence animationClip) {
		// TODO Auto-generated method stub
		
	}
	
	public void attachColliderToBone(Bone bone, ICollider collider){
		BoneCollider boneCollider = new BoneCollider(collider, this, bone.index);
		addCollider(boneCollider);
	}
	
	public Iterator<IKConstraint> getIKChainsIterator(){
		return ikChains.iterator(); 
	}
	
	public void attachIKChain(IKConstraint chain){
		ikChains.add(chain);
	}
	
	public IKConstraint findIKChainWithId(int id){
		for( IKConstraint ikChain : ikChains ){
			if( ikChain.id == id ){
				return ikChain; 
			}
		}
		
		return null; 
	}
	
	public IKConstraint findIKChainWithTag(String tag){
		for( IKConstraint ikChain : ikChains ){
			if( ikChain.tag.equalsIgnoreCase(tag) ){
				return ikChain; 
			}
		}
		
		return null; 
	}
	
	public boolean removeIKChain(IKConstraint ikChain){
		return ikChains.remove(ikChain);
	}
}
