/**
 * CControllerParser.java (NoFrills3D)
 * Created by Joshua Newnham on Apr 23, 2013.
 * Copyright (c) 2013 We Make Play All rights reserved.
 */
package com.wmp.nofrills3d.entities.loaders.collada;

import java.io.IOException;
import java.util.List;
import java.util.StringTokenizer;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;

import com.wmp.nofrills3d.entities.loaders.collada.COLLADAModelLoader.CDataTypeEnum;
import com.wmp.nofrills3d.entities.loaders.collada.COLLADAModelLoader.CSourceSemantic;
import com.wmp.nofrills3d.utils.Logger;

public class CControllerParser extends CBaseParser {

	public static final String TAG = "CControllerParser";
	
	protected enum ParserStateEnum{
		Undefined, 
		Processing_BindShapeMatrix, 
		Processing_Source,
		Processing_Joints, 
		Processing_VertexWeights		
	}

	protected XmlPullParser xpp = null;
	
	protected List<CGeometry> geometryLibrary = null;
	protected CGeometry associatedGeometry = null; 
	
	protected String currentControllerId = "";
	protected String currentControllerName = "";
	
	protected CSkin currentSkin = null;
	protected CSource currentSource = null; 
	
	protected ParserStateEnum processingState = ParserStateEnum.Undefined; 
	
	public static CControllerParser parse(XmlPullParser xpp, List<CGeometry> geometryLibrary){
		return new CControllerParser(xpp, geometryLibrary).start(); 
	}
	
	private CControllerParser(XmlPullParser xpp, List<CGeometry> geometryLibrary){
		this.xpp = xpp; 
		this.geometryLibrary = geometryLibrary; 
	}
	
	public CControllerParser start(){
		if( !xpp.getName().equals(CConstants.TAG_LIBRARY_CONTROLLERS) ){
			return this; 
		}
		
		try{
			while(true){
				xpp.next(); 
				int eventType = xpp.getEventType();
				
				if (eventType == XmlPullParser.END_TAG) {
					if( xpp.getName().equals(CConstants.TAG_LIBRARY_CONTROLLERS) ){
						break; 
					} else{
						parseEndTag(xpp);
					}
				} 
				else if( eventType == XmlPullParser.START_TAG ){
					parseStartTag(xpp);
				}
				else if( eventType == XmlPullParser.TEXT ){
					parseText(xpp);
				}
			}
		
		} catch( XmlPullParserException e ){
			
		} catch( IOException e ){
			
		}
		
		return this; 
	}
	
	protected void parseStartTag(XmlPullParser xpp) throws XmlPullParserException, IOException{
		String tag = xpp.getName(); 
		
		if( tag.equals(CConstants.TAG_CONTROLLER)){
			for( int i=0; i<xpp.getAttributeCount(); i++ ){
				String attr = xpp.getAttributeName(i);
				String val = xpp.getAttributeValue(i);
				
				if( attr.equals(CConstants.ATTR_ID)){
					currentControllerId = val; 
				} else if( attr.equals(CConstants.ATTR_NAME)){
					currentControllerName = val; 
				}
			}
		} else if( tag.equals(CConstants.TAG_SKIN)){
			for( int i=0; i<xpp.getAttributeCount(); i++ ){
				String attr = xpp.getAttributeName(i);
				String val = xpp.getAttributeValue(i);
				
				if( attr.equals(CConstants.ATTR_SOURCE)){
					associatedGeometry = findGeometryWithId(val);
					
					if( associatedGeometry != null ){
						if( associatedGeometry.skin != null ){
							throw new IOException("Trying to assign 2 skins to a geometry object - please ensure that only one skin is associated to each model");
						}
						associatedGeometry.skin = new CSkin(); 
						currentSkin = associatedGeometry.skin; 
						currentSkin.id = currentControllerId; 
						currentSkin.name = currentControllerName;
					}
				}				
			}
		} else if( tag.equals(CConstants.TAG_BIND_SHAPE_MATRIX)){
			processingState = ParserStateEnum.Processing_BindShapeMatrix;
			
			if( currentSkin != null ){
				xpp.next(); 				
				parseDelimitedMatrix(currentSkin.bindShapeMatrix, xpp.getText());
			}
		} else if( tag.equals(CConstants.TAG_SOURCE)){
			processingState = ParserStateEnum.Processing_Source;
			
			currentSource = new CSource(); 
			
			for( int i=0; i<xpp.getAttributeCount(); i++ ){
				String attr = xpp.getAttributeName(i);
				String val = xpp.getAttributeValue(i);
				
				if( attr.equals(CConstants.ATTR_ID)){
					currentSource.id = val; 
				}				
			}
		} else if( tag.equals(CConstants.TAG_SOURCE_FLOAT_ARRAY )){
			if( currentSource != null ){
				for( int i=0; i<xpp.getAttributeCount(); i++ ){
					String attr = xpp.getAttributeName(i);
					String val = xpp.getAttributeValue(i);
					
					if( attr.equals(CConstants.ATTR_ID)){
						 // ignore 
					} else if( attr.equals(CConstants.ATTR_COUNT)){
						 currentSource.length = Integer.parseInt(val);
						 currentSource.dataAsFloat = new float[currentSource.length];
						 currentSource.dataType = CDataTypeEnum.Float;
					}
				}
				
				xpp.next(); 
				
				parseDelimitedFloatArrayForSource(currentSource, xpp.getText());
			}						
		} else if( tag.equals(CConstants.TAG_SOURCE_NAME_ARRAY)){
			if( currentSource != null ){
				for( int i=0; i<xpp.getAttributeCount(); i++ ){
					String attr = xpp.getAttributeName(i);
					String val = xpp.getAttributeValue(i);
					
					if( attr.equals(CConstants.ATTR_ID)){
						 // ignore 
					} else if( attr.equals(CConstants.ATTR_COUNT)){
						 currentSource.length = Integer.parseInt(val);
						 currentSource.dataAsString = new String[currentSource.length];
						 currentSource.dataType = CDataTypeEnum.Name;
					}
				}
				
				xpp.next(); 
				
				parseDelimitedStringArrayForSource(currentSource, xpp.getText());
			}	
		} else if( tag.equals(CConstants.TAG_SOURCE_ACCESSOR)){
			if( currentSource != null ){
				for( int i=0; i<xpp.getAttributeCount(); i++ ){
					String attr = xpp.getAttributeName(i);
					String val = xpp.getAttributeValue(i);
					
					if( attr.equals(CConstants.ATTR_COUNT)){
						currentSource.count = Integer.parseInt(val); 
					} else if( attr.equals(CConstants.ATTR_STRIDE)){
						currentSource.stride = Integer.parseInt(val);
						currentSource.params = new CSourceParam[1];
					}				
				}
			}					
		} else if( tag.equals(CConstants.TAG_SOURCE_ACCESSOR_PARAM)){
			if( currentSource != null ){				
				CSourceParam param = new CSourceParam(); 
				
				for( int i=0; i<xpp.getAttributeCount(); i++ ){
					String attr = xpp.getAttributeName(i);
					String val = xpp.getAttributeValue(i);
					
					if( attr.equals(CConstants.ATTR_NAME)){
						param.name = val;  
					} else if( attr.equals(CConstants.ATTR_TYPE)){
						if( val.equals(CConstants.VALUE_PARAM_FLOAT4x4) ){
							param.type = CDataTypeEnum.Float4x4; 
						} 
					}				
				}
				
				currentSource.params[currentSource.nextParamIndex] = param;
				currentSource.nextParamIndex++;
				
				if( currentSource.nextParamIndex >= currentSource.params.length ){
					CSourceParam[] newParamsArray = new CSourceParam[currentSource.params.length+1];
					System.arraycopy(currentSource.params, 0, newParamsArray, 0, currentSource.params.length);
					currentSource.params = newParamsArray; 
				}
			}					
		} else if( tag.equals(CConstants.TAG_JOINTS) ){
			processingState = ParserStateEnum.Processing_Joints;
			
		} else if( tag.equals(CConstants.TAG_VERTEX_WEIGHTS) ){
			processingState = ParserStateEnum.Processing_VertexWeights; 
			if( currentSkin != null ){				
				parseVertexWeights(xpp);
			}
			
		} else if( tag.equals(CConstants.TAG_INPUT) ){
			if( currentSkin != null ){
				if( processingState == ParserStateEnum.Processing_Joints ){
					String semantic = null; 
					String source = null;
					
					for( int i=0; i<xpp.getAttributeCount(); i++ ){
						String attr = xpp.getAttributeName(i);
						String val = xpp.getAttributeValue(i);
						
						if( attr.equals(CConstants.ATTR_SEMANTIC)){
							semantic = val; 
						} else if( attr.equals(CConstants.ATTR_SOURCE)){
							source = val; 
						}				
					}
					
					if( semantic != null && source != null ){
						CSource cSource = currentSkin.findSourceWithId(source);
						
						if( cSource != null ){
							cSource.type = parseStringToSemantic(semantic);
						}
					}					
				} 
			}
		}
		
	}
	
	protected void parseVertexWeights(XmlPullParser xpp) throws XmlPullParserException, IOException{
		
		if( !xpp.getName().equals(CConstants.TAG_VERTEX_WEIGHTS)){
			return; 
		}
		
		for( int i=0; i<xpp.getAttributeCount(); i++ ){
			
			String attr = xpp.getAttributeName(i);
			String val = xpp.getAttributeValue(i);
			
			if( attr.equals(CConstants.ATTR_COUNT)){
				int len = Integer.parseInt(val); 
				currentSkin.vertexBoneInfluences = new CVertexBoneInfluence[len];
				
				if( associatedGeometry.findSourceOfType(CSourceSemantic.Vertex).count != len){
					Logger.LogWarning(TAG, "Vertex length and vertex weight length do not match");
				}
			}				
		}
				
		// assuming that two inputs are used to define vertex weights; the first being the joint index and the second being the weight (2 strides)
		// (if this is not the case with later version of Collada then this should be abstracted out to a VertexWeight Class as we have done with the PolyList)
		final int JOINT_OFFSET = 0; 
		final int WEIGHT_OFFSET = 1;
		final int STRIDE = 2;
		int inputCount = 0; 
		
		CSource jointSource = null; 		
		CSource weightSource = null;
		
		while(true){
			xpp.next(); 
			int eventType = xpp.getEventType();
			
			if (eventType == XmlPullParser.END_TAG) {
				if( xpp.getName().equals(CConstants.TAG_VERTEX_WEIGHTS) ){
					break;
				}
			} else if( eventType == XmlPullParser.START_TAG ){
				String tag = xpp.getName(); 
				
				if( tag.equals(CConstants.TAG_INPUT) ){
					inputCount++; 
					
					if( inputCount > STRIDE ){
						Logger.LogWarning(TAG, "(parseVertexWeights) Have found more than 2 inputs whilst parsing the vertex weights - review dae file");
					}
					
					String semantic = null; 
					String source = null;
					int offset = 0; 
					
					for( int i=0; i<xpp.getAttributeCount(); i++ ){
						String attr = xpp.getAttributeName(i);
						String val = xpp.getAttributeValue(i);
						
						if( attr.equals(CConstants.ATTR_SEMANTIC)){
							semantic = val; 
						} else if( attr.equals(CConstants.ATTR_SOURCE)){
							source = val; 
						} else if( attr.equals(CConstants.ATTR_OFFSET)){
							offset = Integer.parseInt(val);
						}
					}
					
					if( semantic != null && source != null ){
						CSource cSource = currentSkin.findSourceWithId(source);
						
						if( cSource != null ){
							cSource.type = parseStringToSemantic(semantic);
							
							if( cSource.type == CSourceSemantic.Joint ){
								jointSource = cSource; 
								if( offset != JOINT_OFFSET ){
									Logger.LogWarning(TAG, "(parseVertexWeights) Invalid joint offset (expecting 0)");
								}
							} else if( cSource.type == CSourceSemantic.Weight ){
								weightSource = cSource;
								if( offset != WEIGHT_OFFSET ){
									Logger.LogWarning(TAG, "(parseVertexWeights) Invalid weight offset (expecting 1)");
								}
							}
						}
					}
				} else if( tag.equals(CConstants.TAG_VCOUNT) ){
					xpp.next(); 					
					parseDelimitedIntegerArrayForVertexWeightCounts(currentSkin, xpp.getText());
				} else if( tag.equals(CConstants.TAG_V) ){
					xpp.next(); 					
					parseDelimitedVertexJointAndWeights(currentSkin, xpp.getText(), jointSource, weightSource);
				}
			}
		}
	}
	
	/*
	 * expecting inputs: 
	 * <input semantic="JOINT" source="#Armature_Ape-skin-joints" offset="0"/>
     * <input semantic="WEIGHT" source="#Armature_Ape-skin-weights" offset="1"/>
	 */
	protected void parseDelimitedVertexJointAndWeights(CSkin skin, String data, CSource jointSource, CSource weightSource){
		StringTokenizer parts = new StringTokenizer(data, " ");
		int numTokens = parts.countTokens();
		
		if (numTokens == 0)
			return;
		
		int vertexWeightIndex = 0; 
		int vertexWeightCount = 0; 
		
		float totalWeight = 0; 
		
		for( int i=0; i<numTokens; i+=2 ){			
			int jointIndex = Integer.parseInt(parts.nextToken());
			int weightIndex = Integer.parseInt(parts.nextToken());
			
			float weight = weightSource.dataAsFloat[weightIndex]; 
			totalWeight += weight; 
			
			if( vertexWeightCount < skin.vertexBoneInfluences[vertexWeightIndex].limit ){
				skin.vertexBoneInfluences[vertexWeightIndex].boneIndex[vertexWeightCount] = jointIndex;											
				skin.vertexBoneInfluences[vertexWeightIndex].weights[vertexWeightCount] = weight;				
			} else{
				Logger.LogWarning(TAG, "parseDelimitedVertexJointAndWeights; jointIndex " + jointIndex + " is associated to more vertices that the set limit; removing the lowest weighted bone associated");
				
				int minIndex = -1; 
				float minWeight = Float.MAX_VALUE; 
				
				for( int j=0; j<skin.vertexBoneInfluences[vertexWeightIndex].limit; j++ ){
					if( skin.vertexBoneInfluences[vertexWeightIndex].weights[j] < minWeight ){
						minWeight = skin.vertexBoneInfluences[vertexWeightIndex].weights[j];
						minIndex = j; 
					}									
				}
				
				if( minIndex != -1 && minWeight < weight ){					
					Logger.LogWarning(TAG, "parseDelimitedVertexJointAndWeights; swapping " + minIndex + " with weight " + minWeight + " with " + jointIndex + " with weight " + weight );
					totalWeight -= minWeight; 
					skin.vertexBoneInfluences[vertexWeightIndex].boneIndex[minIndex] = jointIndex; 
					skin.vertexBoneInfluences[vertexWeightIndex].weights[minIndex] = weight;
				}
			}
			
			if( ++vertexWeightCount == skin.vertexBoneInfluences[vertexWeightIndex].actualCount ){
				
				if( Math.abs(totalWeight-1.0f) > 0.05f ){
					Logger.LogWarning(TAG, 
							String.format("parseDelimitedVertexJointAndWeights; total weight for  %d is not equal to 1 (%f)", vertexWeightIndex, totalWeight) );
				}
				
				/*if( totalWeight != 1.0f ){
					if( skin.vertexBoneInfluences[vertexWeightIndex].actualCount == 1 ){
						skin.vertexBoneInfluences[vertexWeightIndex].weights[vertexWeightCount] = 1.0f;
					} else{
						float normalizeFactor = 1.0f / totalWeight; 
						for( int j=0; j<Math.min(skin.vertexBoneInfluences[vertexWeightIndex].actualCount, skin.vertexBoneInfluences[vertexWeightIndex].limit); j++ ){
							skin.vertexBoneInfluences[vertexWeightIndex].weights[j] *= normalizeFactor; 
						}
					}
				}*/
				
				vertexWeightIndex++; 
				vertexWeightCount = 0;
				totalWeight = 0.0f; 
			}			
		}
	}
	
	protected void parseDelimitedIntegerArrayForVertexWeightCounts(CSkin skin, String data){
		StringTokenizer parts = new StringTokenizer(data, " ");
		int numTokens = parts.countTokens();
		
		if (numTokens == 0)
			return;
		
		for( int i=0; i<numTokens; i++ ){
			skin.vertexBoneInfluences[i] = new CVertexBoneInfluence();
			
			String val = parts.nextToken(); 
			if( val != null ){
				int count = Integer.parseInt(val);
				skin.vertexBoneInfluences[i].actualCount = count; 	
			}							
		}		
	}
	
	protected void parseText(XmlPullParser xpp){
		
	}
	
	protected void parseEndTag(XmlPullParser xpp){
		String tag = xpp.getName();
		
		if( tag.equals(CConstants.TAG_SOURCE )){						
			if( currentSource != null ){
				if( currentSkin != null ){
					currentSkin.sources.add(currentSource);
				}
				
				currentSource = null; 
			}
		} 
		
		if( tag.equals(CConstants.TAG_JOINTS)
				|| tag.equals(CConstants.TAG_SOURCE) 
					|| tag.equals(CConstants.TAG_VERTEX_WEIGHTS) 
						|| tag.equals(CConstants.TAG_SOURCE) 
							|| tag.equals(CConstants.TAG_BIND_SHAPE_MATRIX) ){
			processingState = ParserStateEnum.Undefined; 			
		}
	}
	
	protected CGeometry findGeometryWithId(String id){
		id = id.replace("#", "");
		//id = id.trim();
		//id = id.replace(" ", "");
		id = id.replaceAll("\\s", "");
		
		for(CGeometry geometry : geometryLibrary ){
			//if( geometry.id.equals(id)){
			if( stringEqual(id, geometry.id)){
				return geometry; 
			}
		}
		
		return null; 
	}		
	
	protected boolean stringEqual( String a, String b){
		if( a == null || a == null ){
			return false; 
		}
		
		if( a.length() != b.length() ){
			return false;
		}
		
		char[] aChars = a.toCharArray(); 
		char[] bChars = b.toCharArray(); 
		
		for( int i=0; i<aChars.length; i++ ){
			if( aChars[i] != bChars[i] ){
				return false; 
			}
		}
		
		return true; 
	}
}
