/**
 * COLLADAModelLoader.java (NoFrills3D)
 * Created by Joshua Newnham on Apr 15, 2013.
 * Copyright (c) 2013 We Make Play All rights reserved.
 */
package com.wmp.nofrills3d.entities.loaders.collada;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.List;
import java.util.Vector;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;

import android.content.Context;
import android.content.res.Resources;
import android.opengl.Matrix;
import android.util.Log;

import com.wmp.colourexplosion.R;
import com.wmp.nofrills3d.animations.AnimationClip;
import com.wmp.nofrills3d.animations.Keyframe;
import com.wmp.nofrills3d.entities.loaders.ModelLoader;
import com.wmp.nofrills3d.geometry.Bone;
import com.wmp.nofrills3d.geometry.ColladaSkinnedMesh;
import com.wmp.nofrills3d.geometry.IMesh;
import com.wmp.nofrills3d.geometry.Mesh;
import com.wmp.nofrills3d.geometry.SkinnedMesh;
import com.wmp.nofrills3d.geometry.VertexDefinition;
import com.wmp.nofrills3d.material.Material;
import com.wmp.nofrills3d.material.SkinnedMaterial;
import com.wmp.nofrills3d.math.Matrix4;
import com.wmp.nofrills3d.math.Vector3;
import com.wmp.nofrills3d.physics.BoxCollider;
import com.wmp.nofrills3d.scene.Constants;
import com.wmp.nofrills3d.utils.Logger;

public class COLLADAModelLoader extends ModelLoader{	
	
	public static final String TAG = "COLLADAModelLoader";
	
	public enum CLibraryStateEnum{
		Undefined, 
		Geometry		
	}
	
	public enum CDataTypeEnum{
		Undefined,
		Float, 
		Int, 
		Float4x4, 
		Name
	}
	
	public enum CSourceSemantic{
		Undefined, 
		Vertex, 
		Normal, 
		TexCoord, 
		Joint, 
		InvBindMatrix, 
		Weight, 
		Input, 
		Output, 
		Interpolation
	}
	
	protected List<ColladaSkinnedMesh> meshes = new Vector<ColladaSkinnedMesh>(); 
	protected List<CGeometry> geometryLibrary = new Vector<CGeometry>(1); 
	protected List<Bone> visualScenesLibrary = new Vector<Bone>(1); 
	protected List<CAnimation> animationLibrary = new Vector<CAnimation>(); 
	
	/**
	 * @param context
	 */
	public COLLADAModelLoader(Context context) {
		super(context);
		// TODO Auto-generated constructor stub
	}

	/* (non-Javadoc)
	 * @see com.wmp.nofrills3d.entities.loaders.ModelLoader#parseResource(int)
	 */
	@Override
	protected void parseResource(int... resId) {
		InputStream inputStream = null; 
		InputStreamReader inputStreamReader = null; 
		
		try {
			inputStream = context.getResources().openRawResource(resId[0]);
			inputStreamReader = new InputStreamReader(inputStream);
			parseXml(inputStreamReader);
			
			resourcesLoaded = true; 
		} catch( Resources.NotFoundException e ){
			
		} finally{
			if( inputStream != null ){
				try{ inputStream.close(); } catch( Exception e){}
			}
		}
		
		
	}	
	
	public IMesh[] buildAndFetchGeometry(){
		if( !resourcesLoaded ){
			return null; 
		}
		for( CGeometry geometry : geometryLibrary ){
			ColladaSkinnedMesh mesh = createSkinnedMeshFromGeometry(geometry);
			if( mesh != null ){
				meshes.add(mesh);
			}
		}
		
		createAnimations(meshes, animationLibrary);
		
		return getParsedMeshes(); 
	}
	
	protected IMesh[] getParsedMeshes(){
		SkinnedMesh[] res = new SkinnedMesh[meshes.size()];
		for( int i=0; i<meshes.size(); i++ ){
			res[i] = meshes.get(i);
		}
		
		return res; 
	}
	
	protected int parseXml(InputStreamReader reader){
		try{
			XmlPullParserFactory factory = XmlPullParserFactory.newInstance();
			factory.setNamespaceAware(true);
			XmlPullParser xpp = factory.newPullParser();
			
			xpp.setInput(reader); 
			
			int eventType = xpp.getEventType();
			while (eventType != XmlPullParser.END_DOCUMENT) {
				if (eventType == XmlPullParser.START_DOCUMENT) {

				} else if (eventType == XmlPullParser.START_TAG) {
					parseTag(xpp); 
				} else if (eventType == XmlPullParser.END_TAG) {

				} else if (eventType == XmlPullParser.TEXT) {

				}
				eventType = xpp.next();
			}
		}
		catch( XmlPullParserException e ){
			return -2; 
		} catch( IOException e){
			return -1; 
		}
		
		return 0; 
	}
	
	protected void parseTag(XmlPullParser xpp){
		if( xpp.getName().equals(CConstants.TAG_LIBRARY_GEOMETRIES) ){
			CGeometriesParser.parse(xpp, geometryLibrary);
		} else if( xpp.getName().equals(CConstants.TAG_LIBRARY_CONTROLLERS)){
			CControllerParser.parse(xpp, geometryLibrary);
		} else if( xpp.getName().equals(CConstants.TAG_LIBRARY_VISUAL_SCENES)){
			CVisualSceneParser.parse(xpp, visualScenesLibrary);
		} else if( xpp.getName().equals(CConstants.TAG_LIBRARY_ANIMATIONS) ){
			CAnimationParser.parse(xpp, animationLibrary);								
		}
	}
	
	protected ColladaSkinnedMesh createSkinnedMeshFromGeometry(CGeometry geometry){
		ColladaSkinnedMesh mesh = null; 
		
		if( !isValidGeometry(geometry) ){
			return null; 
		}
		
		boolean hasBones = false; 
		
		if( assignRootBoneWithGeometriesSkin(geometry) ){ 
			hasBones = assignBoneIndexes(geometry); 
		}		
		
		int currentOffset = 0; 
		
		float[] positions = null; 
		int positionComponentCount = 0;
		int positionIndiciesOffset = 0;
		
		float[] normals = null;
		int normalComponentCount = 0;
		int normalIndiciesOffset = 0;
		
		float[] texCoords = null;
		int texCoordsComponentCount = 0;
		int texCoordsIndiciesOffset = 0; 
		 
		CVertexBoneInfluence[] boneInfluences = null; 
		int boneIndexComponentCount = CConstants.BONE_INFLUENCE_LIMIT; 
		int boneIndexComponentOffset = 0; 
		
		int boneWeightComponentCount = CConstants.BONE_INFLUENCE_LIMIT; 
		int boneWeightComponentOffset = 0;
		
		if( geometry.hasSourceOfType(CSourceSemantic.Vertex)){
			positions = geometry.findSourceOfType(CSourceSemantic.Vertex).dataAsFloat;
			positionComponentCount = geometry.findSourceOfType(CSourceSemantic.Vertex).stride;
			positionIndiciesOffset = geometry.polyList.findInputForSourceType(CSourceSemantic.Vertex).offset;
			currentOffset += positionComponentCount;  
		}
		
		if( geometry.hasSourceOfType(CSourceSemantic.Normal)){
			normals = geometry.findSourceOfType(CSourceSemantic.Normal).dataAsFloat;
			normalComponentCount = geometry.findSourceOfType(CSourceSemantic.Normal).stride;
			normalIndiciesOffset = geometry.polyList.findInputForSourceType(CSourceSemantic.Normal).offset;
			currentOffset += normalComponentCount; 
		}
		
		if( geometry.hasSourceOfType(CSourceSemantic.TexCoord)){
			texCoords = geometry.findSourceOfType(CSourceSemantic.TexCoord).dataAsFloat;
			texCoordsComponentCount = geometry.findSourceOfType(CSourceSemantic.TexCoord).stride;
			texCoordsIndiciesOffset = geometry.polyList.findInputForSourceType(CSourceSemantic.TexCoord).offset;
			currentOffset += texCoordsComponentCount;
		}	
		
		if( hasBones ){
			boneInfluences = geometry.skin.vertexBoneInfluences; 
			boneIndexComponentOffset = currentOffset; 
			currentOffset += boneIndexComponentCount;
			boneWeightComponentOffset = currentOffset;
		}
		
		short[] unsortIndicies = geometry.polyList.indexes;
		short unsortIndiciesStride = (short)geometry.polyList.stride();
		int indiciesLength = geometry.polyList.getIndiciesLength();		
		
		float[] verts = new float[indiciesLength * geometry.getStride() + 
		                          (hasBones ? indiciesLength * (boneIndexComponentCount + boneWeightComponentCount) : 0)];
		short[] indicies = new short[indiciesLength];
		
		int unsortIndiciesIdx = 0;
		int vertsIndex = 0; 
		
		for( short i=0; i<indiciesLength; i++ ){
			indicies[i] = i; 
			
			int vIdx = unsortIndicies[unsortIndiciesIdx + positionIndiciesOffset];
			
			for( int j=0; j<positionComponentCount; j++ ){				
				verts[vertsIndex++] = positions[vIdx * positionComponentCount + j];
			}
			
			if( normals != null ){
				int nIdx = unsortIndicies[unsortIndiciesIdx + normalIndiciesOffset];
				for( int j=0; j<normalComponentCount; j++ ){					
					verts[vertsIndex++] = normals[nIdx * normalComponentCount + j];
				}
			}
			
			if( texCoords != null ){
				int tIdx = unsortIndicies[unsortIndiciesIdx + texCoordsIndiciesOffset];
				for( int j=0; j<texCoordsComponentCount; j++ ){					
					verts[vertsIndex++] = texCoords[tIdx * texCoordsComponentCount + j];
				}
			}
			
			if( hasBones ){
				for( int j=0; j<CConstants.BONE_INFLUENCE_LIMIT; j++ ){
					verts[vertsIndex++] = boneInfluences[vIdx].boneIndex[j]; 
				}
				
				for( int j=0; j<CConstants.BONE_INFLUENCE_LIMIT; j++ ){
					verts[vertsIndex++] = boneInfluences[vIdx].weights[j]; 
				}
			}
			
			unsortIndiciesIdx += unsortIndiciesStride;
		}
		
		mesh = new ColladaSkinnedMesh(geometry.name);
		if( hasBones ){
			mesh.bindShapeMatrix = createdMatrixWithParsedArray(geometry.skin.bindShapeMatrix.matrix);
		}
		//mesh.bindShapeMatrix = geometry.skin.bindShapeMatrix.clone();
		mesh.pushVertexDefinition( new VertexDefinition(Constants.ATTRIBUTE_POSITION, positionComponentCount));
		
		if( normals != null ){
			mesh.pushVertexDefinition( new VertexDefinition(Constants.ATTRIBUTE_NORMAL, normalComponentCount));
		}
		if( texCoords != null ){
			mesh.pushVertexDefinition( new VertexDefinition(Constants.ATTRIBUTE_TEX_COORD, texCoordsComponentCount));
		}
		
		if( hasBones ){
			mesh.pushVertexDefinition( new VertexDefinition(Constants.ATTRIBUTE_BONE_INDEX, boneIndexComponentCount));
			mesh.pushVertexDefinition( new VertexDefinition(Constants.ATTRIBUTE_BONE_WEIGHT, boneWeightComponentCount));
		}		
		
		mesh.setIndiciesData(indicies);				
		mesh.setVertexData(verts);
		
		if( hasBones ){
			
			assignRootBoneAndBuildHierarchyForMesh(geometry.rootBone, mesh);
			createBindPoseMatrixArrayForMesh(geometry.rootBone, mesh); 
			createInverseBindPoseMatrixArrayForMesh(geometry, mesh);
			
			if( mesh.bindPose.length != mesh.inverseBindPose.length ){
				Logger.LogWarning(TAG, "Bind Pose and Inverse Bind Pose arrays lengths do not match");
			}
			
			buildBoundingBoxFromVertices(mesh, positions, positionComponentCount);
			buildBonesBoundingBoxFromVertices(mesh, positions, positionComponentCount, geometry.skin.vertexBoneInfluences);			
		} else{
			buildBoundingBoxFromVertices(mesh, positions, positionComponentCount);
		}
		
		return mesh; 
	}	
	
	/**
	 * http://collada.org/public_forum/showthread.php/1277-matrix-and-collada
	 * http://www.learnopengles.com/understanding-opengls-matrices/
	 * http://urbanar.blogspot.co.uk/2011/04/from-homography-to-opengl-modelview.html
	 * @param mat
	 * @return
	 */
	protected Matrix4 createdMatrixWithParsedArray(float[] mat){
		Matrix4 res = new Matrix4();
		return createdMatrixWithParsedArray(mat, res); 
	}
	
	protected Matrix4 createdMatrixWithParsedArray(float[] mat, Matrix4 res){
		
		/**
		 0  1  2  3
		 4  5  6  7
		 8  9 10 11
		 12 13 14 15

		 0  1  2 
		 4  5  6 
		 8  9  10		 		 
		 */
				
		Matrix.transposeM(res.matrix, 0, mat, 0);			
		
		return res; 
	}
	
	protected boolean assignRootBoneWithGeometriesSkin(CGeometry geometry){
		if( geometry.skin == null || geometry.skin.id == null ){
			return false; 
		}
		
		for( Bone bone : visualScenesLibrary ){
			if( bone.url.equals(geometry.skin.id) ){
				geometry.rootBone = bone; 
				return true; 
			}
		}
		
		return false; 				
	}
	
	protected boolean assignRootBoneAndBuildHierarchyForMesh(Bone rootBone, SkinnedMesh mesh){		
		mesh.rootBone = rootBone;
		
		int count = rootBone.size();
		
		mesh.boneHierarchy = new int[count];
		mesh.boneHierarchy[0] = 0; 
		
		for( int i=1; i<count; i++ ){
			mesh.boneHierarchy[i] = rootBone.findBoneAtIndex(i).parent.index; 
		}
		
		return true; 
	}
	
	protected boolean createBindPoseMatrixArrayForMesh(Bone rootBone, SkinnedMesh mesh){		
		if( rootBone == null ){
			return false; 
		}
		
		int count = rootBone.size();
		mesh.bindPose = new Matrix4[count];
		
		for( int i=0; i<count; i++ ){
			Bone bone = rootBone.findBoneAtIndex(i);
			
			if( bone == null ){
				Logger.LogWarning(TAG, "createBindPoseMatrixArrayForMesh :: unable to find bone with index " + i);
				continue; 
			}
			
			mesh.bindPose[i] = createdMatrixWithParsedArray(bone.localBonetransform.matrix);
			bone.localBonetransform.set(mesh.bindPose[i].matrix);
		}
		
		return true; 
	}
	
	protected boolean createInverseBindPoseMatrixArrayForMesh(CGeometry geometry, SkinnedMesh mesh){		
		// bind inverse poses 
		CSource invBindPoseSource = geometry.skin.findSourceOfType(CSourceSemantic.InvBindMatrix);
		
		if( invBindPoseSource == null ){
			return false; 
		}
				
		mesh.inverseBindPose = new Matrix4[invBindPoseSource.count];
		
		Matrix4 tmpMat = new Matrix4();
		
		for( int i=0; i<invBindPoseSource.count; i++ ){
			//mesh.inverseBindPose[i] = Matrix4.createIdentity(); 
			
			int offset = i * invBindPoseSource.stride;
			
			for( int j=0; j<invBindPoseSource.stride; j++ ){
				tmpMat.matrix[j] = invBindPoseSource.dataAsFloat[offset + j];  
			}
			
			mesh.inverseBindPose[i] = createdMatrixWithParsedArray(tmpMat.matrix);
		}		
		
		return true; 
	}
	
	protected boolean assignBoneIndexes(CGeometry geometry){
		if( geometry.skin == null || geometry.rootBone == null ){
			return false; 
		}				
		
		CSource boneSource = geometry.skin.findSourceOfType(CSourceSemantic.Joint);
		
		if( boneSource == null || boneSource.dataAsString == null ){
			return false; 
		}
		
		int missingBoneCounter = 0; 
		
		for( int boneIndex = 0; boneIndex<boneSource.length; boneIndex++ ){
			Bone bone = geometry.rootBone.findBoneWithId(boneSource.dataAsString[boneIndex]);
			if( bone != null ){
				bone.index = boneIndex; 
			} else{
				missingBoneCounter++; 
				Logger.LogWarning(TAG, "assignBoneIndexes; Unable to find assocated bone with SID " + boneSource.dataAsString[boneIndex] + " for index " + boneIndex );
			}
		}
		
		return missingBoneCounter == 0; 
	}
	
	protected boolean isValidGeometry(CGeometry geometry){
		if( geometry.sources.size() == 0 || geometry.polyList == null ){
			return false; 
		} else{
			return true; 
		}
	}
	
	protected void createAnimations(final List<ColladaSkinnedMesh> meshes, final List<CAnimation> animationLibrary){
		if( animationLibrary == null || animationLibrary.size() == 0 ){
			return; 
		}
		
		float[] tmpTransform = new float[16]; 
		
		for(CAnimation animation : animationLibrary ){
			SkinnedMesh associatedMesh = null;
			AnimationClip animationClip = null; 
			
			for(SkinnedMesh mesh : meshes ){
				if( mesh.rootBone != null && mesh.rootBone.findBoneWithId(animation.bone) != null ){
					associatedMesh = mesh; 
					
					if( associatedMesh.animationClips == null ){
						associatedMesh.animationClips = new Vector<AnimationClip>(1); 
						animationClip = new AnimationClip();
						associatedMesh.animationClips.add(animationClip);
					} else{
						animationClip = associatedMesh.animationClips.get(0); // only supports one animation clip  
					}										
					
					break; 
				}
			}
			
			if( associatedMesh == null ){
				Logger.LogError(TAG, "createAnimations; Unable to find assocated mesh that owns the bone " + animation.bone);
				continue; 
			}
			
			Bone bone = associatedMesh.rootBone.findBoneWithId(animation.bone);
			if( bone == null ){
				Logger.LogError(TAG, "createAnimations; Unable to find assocated bone with the id " + animation.bone);
				continue; 
			}
			
			CSource timeSource = animation.timeSource; 
			CSource transformSource = animation.transformSource; 
			
			if( timeSource == null || transformSource == null || timeSource.count != transformSource.count ){
				Logger.LogError(TAG, "Invalid source data for animation " + animation.id);
				continue;
			}
			
			Matrix4 tmpMat = new Matrix4(); 
			
			int count = transformSource.count; 
			
			for( int i=0; i<count; i++ ){
				float time = timeSource.dataAsFloat[i]; 				
				
				for( int j=0; j<16; j++ ){
					tmpTransform[j] = transformSource.dataAsFloat[i*16+j];
				}
				
				float[] transform = new float[16];
				//Matrix.transposeM(transform, 0, tmpTransform, 0);
				createdMatrixWithParsedArray(tmpTransform, tmpMat);
				
				for( int j=0; j<16; j++ ){
					transform[j] = tmpMat.matrix[j]; 
				}
				
				//if( time > animationClip.duration ){
				//	animationClip.duration = time; 
				//}							
				
				animationClip.addKeyframe( new Keyframe(bone, transform, time) );
			}
			
		}				
		
		for(SkinnedMesh mesh : meshes ){
			if( mesh.animationClips != null ){
				mesh.animationClips.get(0).sortKeyframes(); 				
			}
		}							
	}
	
	protected void buildBoundingBoxFromVertices(Mesh mesh, float[] verts, int stride){
		mesh.boundingBox = new BoxCollider(null, new Vector3(Float.MAX_VALUE), new Vector3(Float.MIN_VALUE));
		
		for( int i=0; i<verts.length; i+=stride ){
			float x = verts[i]; 
			float y = verts[i+1];
			float z = (stride == 3 ? verts[i+2] : 0);
			
			if( x < mesh.boundingBox.min.x ){
				mesh.boundingBox.min.x = x; 
			} 
			if( x > mesh.boundingBox.max.x ){
				mesh.boundingBox.max.x = x;
			}
			
			if( y < mesh.boundingBox.min.y ){
				mesh.boundingBox.min.y = y; 
			} 
			if( y > mesh.boundingBox.max.y ){
				mesh.boundingBox.max.y = y;
			}
			
			if( stride == 3 ){
				if( z < mesh.boundingBox.min.z ){
					mesh.boundingBox.min.z = z; 
				}
				if( z > mesh.boundingBox.max.z ){
					mesh.boundingBox.max.z = z;
				}
			}
		}
	}
	
	protected void buildBonesBoundingBoxFromVertices(SkinnedMesh mesh, float[] verts, int stride, CVertexBoneInfluence[] boneInfluences){
		int boneCount = mesh.boneHierarchy.length; 
		
		mesh.boneBoundingBox = new BoxCollider[boneCount];
		
		int boneInfluenceIndex = 0; 
		
		for( int i=0; i<verts.length; i+=stride ){
			float x = verts[i]; 
			float y = verts[i+1];
			float z = (stride == 3 ? verts[i+2] : 0); 						
			
			if( boneInfluenceIndex >= boneInfluences.length ){
				if( Logger.Logging ){
					Logger.LogWarning(TAG, "buildBonesBoundingBoxFromVertices; indexing incorrect boneInfluenceIndex = " + boneInfluenceIndex + ", i = " + i
							+ ", boneInfluences.length = " + boneInfluences.length + ", verts.length = " + verts.length);
				}
				return; 
			}
			CVertexBoneInfluence boneInfluence = boneInfluences[boneInfluenceIndex++];
						
			// for each bone
			int[] boneIndexes = boneInfluence.boneIndex; 
			for( int j=0; j<boneIndexes.length; j++ ){
				if( mesh.boneBoundingBox[j] == null ){
					mesh.boneBoundingBox[j] = new BoxCollider(null, new Vector3(Float.MAX_VALUE), new Vector3(Float.MIN_VALUE));										
				}
				
				if( x < mesh.boneBoundingBox[j].min.x ){
					mesh.boneBoundingBox[j].min.x = x; 
				} else if( x > mesh.boneBoundingBox[j].max.x ){
					mesh.boneBoundingBox[j].max.x = x;
				}
				
				if( y < mesh.boneBoundingBox[j].min.y ){
					mesh.boneBoundingBox[j].min.y = y; 
				} else if( y > mesh.boneBoundingBox[j].max.y ){
					mesh.boneBoundingBox[j].max.y = y;
				}
				
				if( stride == 3 ){
					if( z < mesh.boneBoundingBox[j].min.z ){
						mesh.boneBoundingBox[j].min.x = z; 
					} else if( z > mesh.boneBoundingBox[j].max.z ){
						mesh.boneBoundingBox[j].max.z = z;
					}
				}
			}
		}
	}
	
	@Override
	public Material getDefaultMaterial(){
		return new SkinnedMaterial(context, R.raw.skinned_basic_texture_dae_vs, R.raw.skinned_basic_texture_dae_fs);
	}
}
