package com.wmp.nofrills3d.entities.loaders.md5;

import java.io.InputStream;
import java.util.List;
import java.util.Vector;

import android.content.Context;

import com.wmp.colourexplosion.R;
import com.wmp.nofrills3d.entities.loaders.ModelLoader;
import com.wmp.nofrills3d.geometry.Bone;
import com.wmp.nofrills3d.geometry.IMesh;
import com.wmp.nofrills3d.geometry.SkinnedMesh;
import com.wmp.nofrills3d.geometry.VertexDefinition;
import com.wmp.nofrills3d.material.Material;
import com.wmp.nofrills3d.material.SkinnedMaterial;
import com.wmp.nofrills3d.math.Matrix4;
import com.wmp.nofrills3d.scene.Constants;
import com.wmp.nofrills3d.utils.Logger;

public class MD5ModelLoader extends ModelLoader{

	public static final String TAG = "MD5ModelLoader";
	
	protected static final boolean DEBUG_INCLUDE_BONES = true; 
	
	protected MD5MeshParser meshParser;	
	
	protected MD5Data md5Data = null; 
	
	public MD5ModelLoader(Context context) {
		super(context);
		
	}

	protected List<SkinnedMesh> meshes = new Vector<SkinnedMesh>();

	@Override
	protected void parseResource(int... resId) {
		meshParser = new MD5MeshParser();				
		
		InputStream is = null; 
		
		try{
			is = context.getResources().openRawResource(resId[0]);
			
			md5Data = meshParser.parse(is);
			
		} catch(Exception e ){
			if( Logger.Logging ){
				Logger.LogError(TAG, e.toString());
			}			
		} finally{
			if( is != null ){
				try{ is.close(); } catch( Exception e ){}
			}
		}								
		
	}
	
	public IMesh[] buildAndFetchGeometry(){
		if( md5Data != null ){
			 buildMesh(md5Data);
			 return null; 
		}
		
		return getParsedMeshes(); 
	}

	@Override
	protected IMesh[] getParsedMeshes() {
		SkinnedMesh[] res = new SkinnedMesh[meshes.size()];
		for( int i=0; i<meshes.size(); i++ ){
			res[i] = meshes.get(i);
		}
		
		return res; 
	} 
	
	protected void buildMesh(MD5Data md5Data){
		Bone[] bones = null;		
		Matrix4[] boneInverseMatrix = null; 				
		
		// create a mesh for each MD5Mesh 
		for( int i=0; i<md5Data.meshes.length; i++ ){
			MD5Mesh md5Mesh = md5Data.meshes[i];
			md5Mesh.calculatePositions(md5Data.joints);
			md5Mesh.calcualteNormals(); 
			
			if( md5Data.numJoints > 0 ){
				boneInverseMatrix = new Matrix4[md5Data.numJoints];
				bones = new Bone[md5Data.numJoints];
				createBones(md5Data, bones); 
			}
			
			Bone rootBone = null; 
			
			if( bones != null ){
				rootBone = findRootBoneForMesh(md5Mesh, md5Data, bones);							
			}
			
			SkinnedMesh mesh = new SkinnedMesh("md5_mesh");
			mesh.rootBone = rootBone;
			
			// build temp bone hierarchy 
			mesh.boneHierarchy = new int[bones.length];
			mesh.boneHierarchy[0] = 0; 
			
			for( int j=1; j<bones.length; j++ ){
				mesh.boneHierarchy[j] = rootBone.findBoneAtIndex(j).parent.index; 
			}
			
			createBindPose(mesh, rootBone);						
				
			// position count + normal count + uv + bone index + weights 
			int stride = 3 + 3 + 2 + (DEBUG_INCLUDE_BONES ? MD5Constants.MAX_BONES_PER_VERTEX + MD5Constants.MAX_BONES_PER_VERTEX : 0);  
				
			int indiciesLength = md5Mesh.triCount * 3; 
			short[] indicies = new short[indiciesLength];
			float[] vertexData = new float[stride * indiciesLength];
			int vertexDataIndex = 0; 
			int k = 0; 
			
			for( short j=0; j<indiciesLength; j++ ){
				indicies[j] = j;
				//indicies[j] = (short)md5Mesh.tris[j];				
				
				int vIdx = md5Mesh.tris[j];
				//int vIdx = j; 
				MD5Vert vert = md5Mesh.verts[vIdx];
				
				for( k=0; k<3; k++ ){
					vertexData[vertexDataIndex++] = vert.position[k]; 
				}
				
				for( k=0; k<3; k++ ){
					vertexData[vertexDataIndex++] = vert.normal[k]; 
				}
				
				for( k=0; k<2; k++ ){
					vertexData[vertexDataIndex++] = vert.uv[k]; 
				}
				
				if( DEBUG_INCLUDE_BONES ){
					for( k=0; k<MD5Constants.MAX_BONES_PER_VERTEX; k++ ){					
						if( k<vert.weightCount ){
							MD5JointWeight weight = md5Mesh.weights[vert.weightIndex+k];
							vertexData[vertexDataIndex++] = weight.jointIndex;						
						} else{
							vertexData[vertexDataIndex++] = 0.0f; 						
						}
					}
					
					
					for( k=0; k<MD5Constants.MAX_BONES_PER_VERTEX; k++ ){					
						if( k<vert.weightCount ){
							MD5JointWeight weight = md5Mesh.weights[vert.weightIndex+k];
							vertexData[vertexDataIndex++] = weight.weight;
						} else{
							vertexData[vertexDataIndex++] = 0.0f; 						
						}
					}					
				}
			}	
			
			mesh.pushVertexDefinition( new VertexDefinition(Constants.ATTRIBUTE_POSITION, 3));			
			mesh.pushVertexDefinition( new VertexDefinition(Constants.ATTRIBUTE_NORMAL, 3));
			mesh.pushVertexDefinition( new VertexDefinition(Constants.ATTRIBUTE_TEX_COORD, 2));	
			if( DEBUG_INCLUDE_BONES ){
				mesh.pushVertexDefinition( new VertexDefinition(Constants.ATTRIBUTE_BONE_INDEX, MD5Constants.MAX_BONES_PER_VERTEX));
				mesh.pushVertexDefinition( new VertexDefinition(Constants.ATTRIBUTE_BONE_WEIGHT, MD5Constants.MAX_BONES_PER_VERTEX));
			}
			
			mesh.setIndiciesData(indicies);
			mesh.setVertexData(vertexData);
			
			meshes.add(mesh);			
		}						
		
	}
	
	/** 
	 * create bind and inverse bind pose 
	 * @param mesh
	 * @param rootBone
	 */
	protected void createBindPose(SkinnedMesh mesh, Bone rootBone){
		int count = rootBone.size();
		mesh.bindPose = new Matrix4[count];
		mesh.inverseBindPose = new Matrix4[count]; 
		
		for( int i=0; i<count; i++ ){
			Bone bone = rootBone.findBoneAtIndex(i);
			
			if( bone == null ){
				Logger.LogWarning(TAG, "createBindPoseMatrixArrayForMesh :: unable to find bone with index " + i);
				continue; 
			}
			
			mesh.bindPose[i] = bone.localBonetransform.clone();
			mesh.inverseBindPose[i] = bone.invBoneTransform.clone(); 
			
			//mesh.inverseBindPose[i] = Matrix4.createIdentity(); 
			//Matrix4.invert(mesh.bindPose[i], mesh.inverseBindPose[i]);
		}
	}
	
	protected Bone findRootBoneForMesh(MD5Mesh md5Mesh, MD5Data md5Data, Bone[] bones){
		MD5JointWeight weight = md5Mesh.weights[0];
		Bone bone = bones[weight.jointIndex];
			
		while( bone.parent != null ){
			bone = bone.parent;
		}
			
		return bone; 						
	}
	
	protected void createBones(MD5Data md5Data, Bone[] bones){
		
		Matrix4[] boneInverseMatrices = new Matrix4[md5Data.numJoints];
		
		for( int i=0; i<md5Data.numJoints; i++ ){
			MD5Joint joint = md5Data.joints[i];
			
			Bone bone = new Bone();
			bone.id = joint.name; 
			bone.sid = joint.name; 
			bone.index = i;
			
			bone.localBonetransform = Matrix4.createFromTranslationAndQuaternion(joint.position[0], joint.position[1], joint.position[2], 
					joint.orientation[0], joint.orientation[1], joint.orientation[2], joint.orientation[3]);					
						
			bone.invBoneTransform = Matrix4.createIdentity(); 
			Matrix4.invert(bone.localBonetransform, bone.invBoneTransform);
			boneInverseMatrices[i] = bone.invBoneTransform.clone(); 
			
			/*
			 * bone transformations are stored in world space so we need to transform world space to local/object space 
			 * pW := W * p
			 * W-1 * pW = W-1 * ( W * p ) = p
			*/
			if( joint.parentIndex != -1 ){													
				Matrix4.multiply(boneInverseMatrices[joint.parentIndex], bone.localBonetransform, bone.localBonetransform);
			}			
			
			bones[i] = bone; 
			
			// assume its order is hierarchy (ie parent has already been added) 
			if( joint.parentIndex != -1 ){
				if( joint.parentIndex > i ){
					if( Logger.Logging ){
						Logger.LogWarning(TAG, "MD5Joints out of order - unable to build bone hierarchy");
					}
				} else{
					Bone parent = bones[joint.parentIndex];
					bone.parent = parent; 
					parent.children.add(bone);
				}
			}
		}			
	}	
	
	@Override
	public Material getDefaultMaterial(){
		return new SkinnedMaterial(context, R.raw.skinned_basic_texture_dae_vs, R.raw.skinned_basic_texture_dae_fs);
	}
}
