/**
 * ObjModelLoader.java (WMP3D)
 * Created by Joshua Newnham on Apr 10, 2013.
 * Copyright (c) 2013 We Make Play All rights reserved.
 */
package com.wmp.nofrills3d.entities.loaders.obj;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;

import android.content.Context;

import com.wmp.nofrills3d.entities.loaders.ModelLoader;
import com.wmp.nofrills3d.geometry.IMesh;
import com.wmp.nofrills3d.geometry.Mesh;
import com.wmp.nofrills3d.geometry.VertexDefinition;
import com.wmp.nofrills3d.scene.Constants;

public class ObjModelLoader extends ModelLoader{

	private final String VERTEX = "v";
	private final String FACE = "f";
	private final String TEXCOORD = "vt";
	private final String NORMAL = "vn";
	private final String OBJECT = "o";
	
	protected ArrayList<Float> verticies;
	protected ArrayList<Float> normals; 
	protected ArrayList<Float> texCoords; 
	
	protected ArrayList<Integer> faces;
	protected ArrayList<Integer> faceNormalIdx;
	protected ArrayList<Integer> faceTexIdx;
	
	protected int vertexCount = 0; 
	protected int normalCount = 0; 
	protected int texCoordsCount = 0; 
	protected int triangleCount = 0; 
	
	protected List<Mesh> meshes = new Vector<Mesh>(); 
	
	protected Mesh currentMesh = null; 
	
	public ObjModelLoader(Context context){
		super(context);
	}
	
	/* (non-Javadoc)
	 * @see com.wmp.wmp3d.entities.loaders.ModelLoader#parseResource(int)
	 */
	@Override
	protected void parseResource(int... resId) {
		InputStream is = null; 
		BufferedReader in = null; 
		String line; 
		
		verticies = new ArrayList<Float>(); 
		normals = new ArrayList<Float>(); 
		texCoords = new ArrayList<Float>(); 	
		
		faces = new ArrayList<Integer>();
		faceNormalIdx = new ArrayList<Integer>(); 
		faceTexIdx = new ArrayList<Integer>(); 

		try{
			is = context.getResources().openRawResource(resId[0]);
			in = new BufferedReader(new InputStreamReader(is));
			
			while ((line = in.readLine()) != null) {
				StringTokenizer parts = new StringTokenizer(line, " ");
				int numTokens = parts.countTokens();
				
				if (numTokens == 0)
					continue;
				
				String type = parts.nextToken();
				
				if( type.equals(VERTEX)){
					verticies.add(Float.parseFloat(parts.nextToken()));
					verticies.add(Float.parseFloat(parts.nextToken()));
					verticies.add(Float.parseFloat(parts.nextToken()));
					
					vertexCount++; 
				}
				else if( type.equals(NORMAL)){
					normals.add(Float.parseFloat(parts.nextToken()));
					normals.add(Float.parseFloat(parts.nextToken()));
					normals.add(Float.parseFloat(parts.nextToken()));
					
					normalCount++; 
				}
				else if( type.equals(TEXCOORD)){
					float u = Float.parseFloat(parts.nextToken());
					float v = Float.parseFloat(parts.nextToken());
					texCoords.add(u);
					texCoords.add(v);
					
					texCoordsCount++; 
				}
				else if( type.equals(FACE)){
										
					for (int j = 0; j < 3; j++) {
						String facePart = parts.nextToken();
						// another tokenizer - based on /
						StringTokenizer faceToken = new StringTokenizer(facePart, "/");
						
						int vert = 0; int texc = 0; int vertN = 0; 
						
						// subtract 1 because they're not 0 based 
						vert = (Integer.parseInt(faceToken.nextToken()) - 1); 
						texc = (Integer.parseInt(faceToken.nextToken()) - 1);
						vertN = (Integer.parseInt(faceToken.nextToken()) - 1);
					
						faces.add(vert);											
						faceTexIdx.add(texc);
						faceNormalIdx.add(vertN);
					}										
					
					triangleCount++; 
				}
				else if( type.equals(OBJECT)){
					if( currentMesh != null ){
						buildCurrentMesh();
						
						if( meshes == null ){
							meshes = new Vector<Mesh>(); 
						}
						
						meshes.add(currentMesh);
					} 
					
					String objName = parts.hasMoreTokens() ? parts.nextToken() : ""; 
					currentMesh = new Mesh(objName);
					
					flushData(); 	
				}
			}
			
		} catch(Exception e){
			
		} finally{
			if( in != null ){
				try{
					in.close();
					in = null; 
				} catch(Exception e){}
			}			
		}	
				
	}	
	
	public IMesh[] buildAndFetchGeometry(){
		
		if( vertexCount > 0 && currentMesh != null ){
			buildCurrentMesh();
			
			if( meshes == null ){
				meshes = new Vector<Mesh>(); 
			}
			
			meshes.add(currentMesh);
			
			return getParsedMeshes(); 
		}
		
		return null; 
	}
	
	protected IMesh[] getParsedMeshes(){
		Mesh[] res = new Mesh[meshes.size()];
		for( int i=0; i<meshes.size(); i++ ){
			res[i] = meshes.get(i);
		}
		
		return res; 
	}
	
	protected void buildCurrentMesh(){
		if( currentMesh == null || vertexCount == 0 ){
			return; 
		}
		
		VertexDefinition vertexDef = new VertexDefinition(Constants.ATTRIBUTE_POSITION, 3);
		VertexDefinition normalDef = new VertexDefinition(Constants.ATTRIBUTE_NORMAL, 3);
		VertexDefinition texCoordsDef = new VertexDefinition(Constants.ATTRIBUTE_TEX_COORD, 2);
		
		currentMesh.pushVertexDefinition(vertexDef, normalDef, texCoordsDef);
		
		// create the arrays for our mesh 				
		short[] indicesArray = new short[faces.size()];
		float[] vertexArray = new float[faces.size() * 8];
		
		int vertexIndex = 0; 
		
		// for each face 
		for( int i=0; i<faces.size(); i++ ){
			int vertIdx = faces.get(i);
			int texcIdx = faceTexIdx.get(i);
			int vertNIdx = faceNormalIdx.get(i);
			
			indicesArray[i] = (short)i;
			
			// vertices positions
			vertexArray[vertexIndex++] = (float)verticies.get(vertIdx * 3 + 0); // position x
			vertexArray[vertexIndex++] = (float)verticies.get(vertIdx * 3 + 1); // position y
			vertexArray[vertexIndex++] = (float)verticies.get(vertIdx * 3 + 2); // position z
			
			// normals 
			vertexArray[vertexIndex++] = (float)normals.get(vertNIdx * 3 + 0); // x
			vertexArray[vertexIndex++] = (float)normals.get(vertNIdx * 3 + 1); // y
			vertexArray[vertexIndex++] = (float)normals.get(vertNIdx * 3 + 2); // z
			
			// uv's 
			vertexArray[vertexIndex++] = (float)texCoords.get(texcIdx * 2 + 0); // u
			vertexArray[vertexIndex++] = (float)texCoords.get(texcIdx * 2 + 1); // v			
		}	
		
		currentMesh.setIndiciesData(indicesArray);
		currentMesh.setVertexData(vertexArray);
	}	
	
	protected void flushData(){
		verticies.clear(); 
		normals.clear(); 
		texCoords.clear(); 
		
		faces.clear();
		faceNormalIdx.clear(); 
		faceTexIdx.clear();
		
		vertexCount = 0; 
		normalCount = 0; 
		texCoordsCount = 0; 
		triangleCount++; 
	}	
}
