/**
 * Mesh.java (ColourExplosion)
 * Created by Joshua Newnham on Apr 9, 2013.
 * Copyright (c) 2013 We Make Play All rights reserved.
 */
package com.wmp.nofrills3d.geometry;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.ShortBuffer;
import java.util.List;
import java.util.Vector;

import android.opengl.GLES20;
import android.text.InputFilter.LengthFilter;

import com.wmp.nofrills3d.material.Material;
import com.wmp.nofrills3d.math.Vector3;
import com.wmp.nofrills3d.physics.BoxCollider;
import com.wmp.nofrills3d.physics.SphereCollider;
import com.wmp.nofrills3d.scene.Constants;
import com.wmp.nofrills3d.utils.Logger;

public class Mesh implements IMesh {
	
	public static final String TAG = "Mesh";	
	
	public String tag = "";
	
	public BoxCollider boundingBox = null; 
	
	public SphereCollider boundingSphere = null; 	
	
	public int glDrawingMode = GLES20.GL_TRIANGLES;
	
	protected int stride = 0; 
	
	protected FloatBuffer vertexBuffer; 
	
	protected int vertexLength = 0; 
	
	protected ShortBuffer indiciesBuffer;
	
	protected int indiciesLength = 0; 
	
	protected List<VertexDefinition> meshDefinition = new Vector<VertexDefinition>(); 
	
	protected boolean attributesBound = false; 
	
	public Mesh(String tag){
		this.tag = tag; 
	}
	
	public String getTag(){
		return tag; 
	}
	
//	protected void bindAttribute(Material material){
//		attributesBound = true;
//		
//		final int[] linkStatus = new int[1]; 
//		
//		for( VertexDefinition elem : meshDefinition ){
//			int locationId = meshDefinition.indexOf(elem);
//			GLES20.glBindAttribLocation(material.getProgram(), locationId, elem.attribute);
//		}
//				
//	}
	
	public void bind(Material material){
//		if( !attributesBound ){
//			bindAttribute(material); 
//		}
		
		for( VertexDefinition elem : meshDefinition ){
			vertexBuffer.position(elem.offset); 
			int attrPtr = material.getAttribLocation(elem.attribute);
			if( attrPtr != -1 ){
				GLES20.glVertexAttribPointer(attrPtr, elem.count, elem.type, false, stride, vertexBuffer);
				GLES20.glEnableVertexAttribArray(attrPtr);
			}
		}
	}
	
	public void draw(){		
		GLES20.glDrawElements(glDrawingMode, indiciesLength, GLES20.GL_UNSIGNED_SHORT, indiciesBuffer);
		checkGlError("glDrawElements");
	}
	
	public void unbind(Material material){				
		for( VertexDefinition elem : meshDefinition ){
			int attrPtr = material.getAttribLocation(elem.attribute);
			if( attrPtr != -1 ){
				GLES20.glDisableVertexAttribArray(attrPtr);
			}
		}			
	}
	
	public void dispose(){
		vertexBuffer = null; 
		indiciesBuffer = null; 
	}
	
	public void setVertexData(float[] data){
		vertexLength = data.length; 
		
		vertexBuffer = ByteBuffer.allocateDirect(data.length * FLOAT_SIZE_BYTES).order(ByteOrder.nativeOrder()).asFloatBuffer(); 		
		vertexBuffer.put(data);
		vertexBuffer.position(0);
		
		buildBoundingBoxFromVertices(data); 
	}
	
	public void setIndiciesData(short[] data){
		indiciesLength = data.length; 
		
		indiciesBuffer = ByteBuffer.allocateDirect(indiciesLength * SHORT_SIZE_BYTES).order(ByteOrder.nativeOrder()).asShortBuffer();
		indiciesBuffer.put(data);
		indiciesBuffer.position(0);
	}
		
	public void pushVertexDefinition(VertexDefinition... vd){
		
		for(VertexDefinition item : vd ){
			// set offset for new vertex definition
			int offset = 0;		
			for( VertexDefinition elem : meshDefinition ){
				offset += elem.count;
			}		
			
			item.offset = offset; 
			
			// update elementStride (in bytes) 
			stride = (offset + item.count) * FLOAT_SIZE_BYTES; 
			
			// add vertex definition to mesh 
			meshDefinition.add(item);
		}					
	}
	
	public VertexDefinition findVertexDefinitionWithName(String name){
		for(VertexDefinition vd : meshDefinition ){
			if( vd.name.equals(name) ){
				return vd; 
			}
		}
		
		return null; 
	}
	
	protected void checkGlError(String op) {
		int error;
		while ((error = GLES20.glGetError()) != GLES20.GL_NO_ERROR) {
			Logger.LogError(TAG, op + ": glError " + error);
			throw new RuntimeException(op + ": glError " + error);
		}
	}
	
	public BoxCollider getBoundingBox(){
		return this.boundingBox; 
	}
	
	protected void buildBoundingBoxFromVertices(float[] data){
		
		VertexDefinition vd = findVertexDefinitionWithName(Constants.ATTRIBUTE_POSITION);
		
		if( vd == null ){
			return; 
		}
		
		int count = vd.count; 
		
		Vector3 min = new Vector3(Float.MAX_VALUE);
		Vector3 max = new Vector3(Float.MIN_VALUE);
		
		Vector3 vPos = Vector3.createZero(); 
		
		for( int i=0; i<vertexLength; i+=stride){
						
			vPos.x = data[i]; 
			if( count >= 2 ) vPos.y = data[i+1];
			if( count == 3 ) vPos.z = data[i+2];
			
			min = Vector3.min(min, vPos);
			max = Vector3.max(max, vPos);
			
		}

		boundingBox = new BoxCollider(null, min, max);
		
	}
}
