/**
 * Mesh.java (ColourExplosion)
 * Created by Joshua Newnham on Apr 9, 2013.
 * Copyright (c) 2013 We Make Play All rights reserved.
 */
package com.wmp.nofrills3d.geometry;

import static android.opengl.GLES20.GL_FLOAT;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;

import android.opengl.GLES20;

import com.wmp.nofrills3d.material.Material;
import com.wmp.nofrills3d.scene.Constants;

public class MeshArray extends Mesh {

	public static final int FLOAT_SIZE_BYTES = 4;
	public static final int INT_SIZE_BYTES = 4;
	public static final int SHORT_SIZE_BYTES = 2;
	
	public static final int POSITION_COMPONENT_COUNT = 3; 
	public static final int NORMAL_COMPONENT_COUNT = 3;
	public static final int COLOUR_COMPONENT_COUNT = 3;
	public static final int TEXTURE_COMPONENT_COUNT = 2; 	
	
	protected String attrPosition = Constants.ATTRIBUTE_POSITION;
	protected String attrNormal = Constants.ATTRIBUTE_NORMAL;
	protected String attrTextureCoords = Constants.ATTRIBUTE_TEX_COORD;
	protected String attrColour = Constants.ATTRIBUTE_COLOUR;
	
	protected FloatBuffer vertexBuffer;
	protected FloatBuffer normalBuffer;
	protected FloatBuffer textureCoordsBuffer;
	protected FloatBuffer colourBuffer; 
	protected IntBuffer indicesBuffer;
	
	protected int vertexCount; 
	protected int normalCount; 
	protected int textureCoordCount; 
	protected int colourCount; 
	protected int indicesCount; 
	
	public MeshArray(String tag){
		super(tag);
	}
	
	public void setTextureCoords(float[] data, String attr){
		attrTextureCoords = attr; 
		setTextureCoords(data);
	}
	
	public void setTextureCoords(float[] data){
		if( textureCoordsBuffer == null ){
			textureCoordsBuffer = ByteBuffer.allocateDirect(data.length * FLOAT_SIZE_BYTES).order(ByteOrder.nativeOrder()).asFloatBuffer();
			textureCoordsBuffer.put(data);
			textureCoordsBuffer.position(0);
		} else{
			textureCoordsBuffer.limit(data.length);
			textureCoordsBuffer.put(data);
			textureCoordsBuffer.position(0);
		}
		
		textureCoordCount = data.length / TEXTURE_COMPONENT_COUNT;				
	}
	
	public void setColourCoords(float[] data, String attr){
		attrColour = attr; 
		setColourCoords(data);
	}
	
	public void setColourCoords(float[] data){
		if( colourBuffer == null ){
			colourBuffer = ByteBuffer.allocateDirect(data.length * FLOAT_SIZE_BYTES).order(ByteOrder.nativeOrder()).asFloatBuffer();
			colourBuffer.put(data);
			colourBuffer.position(0);
		} else{
			colourBuffer.limit(data.length);
			colourBuffer.put(data);
			colourBuffer.position(0);
		}
		
		colourCount = data.length / COLOUR_COMPONENT_COUNT;
	}
	
	public void setNormals(float[] data, String attr){
		attrNormal = attr; 
		setNormals(data);
	}
	
	public void setNormals(float[] data){
		if( normalBuffer == null ){
			normalBuffer = ByteBuffer.allocateDirect(data.length * FLOAT_SIZE_BYTES).order(ByteOrder.nativeOrder()).asFloatBuffer();
			normalBuffer.put(data);
			normalBuffer.position(0);
		} else{
			normalBuffer.limit(data.length);
			normalBuffer.put(data);
			normalBuffer.position(0);
		}
		
		normalCount = data.length / NORMAL_COMPONENT_COUNT;
	}
	
	public void setVertices(float[] data, String attr){
		attrPosition = attr; 
		setVertices(data);
	}
	
	public void setVertices(float[] data){
		if( vertexBuffer == null ){
			vertexBuffer = ByteBuffer.allocateDirect(data.length * FLOAT_SIZE_BYTES).order(ByteOrder.nativeOrder()).asFloatBuffer();
			vertexBuffer.put(data);
			vertexBuffer.position(0);
		} else{
			vertexBuffer.limit(data.length);
			vertexBuffer.put(data);
			vertexBuffer.position(0);
		}
		
		vertexCount = data.length / POSITION_COMPONENT_COUNT;
	}
	
	/*public void setTextureCoords(int[] data){	
		
		if( indicesBuffer == null ){
			indicesBuffer = ByteBuffer.allocateDirect(data.length * FLOAT_SIZE_BYTES).order(ByteOrder.nativeOrder()).asIntBuffer();
			indicesBuffer.put(data);
			indicesBuffer.position(0);
		} else{
			indicesBuffer.limit(data.length);
			indicesBuffer.put(data);
			indicesBuffer.position(0);
		}
		
		indicesCount = data.length;				
	}*/
	
	@Override
	public void bind(Material material){
		
		if( vertexBuffer != null ){
			GLES20.glVertexAttribPointer(material.getAttribLocation(attrPosition), POSITION_COMPONENT_COUNT, GL_FLOAT, false, 0, vertexBuffer);
			GLES20.glEnableVertexAttribArray(material.getAttribLocation(attrPosition));		        
			vertexBuffer.position(0);
		}
		
		if( normalBuffer != null ){
			GLES20.glVertexAttribPointer(material.getAttribLocation(attrNormal), NORMAL_COMPONENT_COUNT, GL_FLOAT, false, 0, normalBuffer);
			GLES20.glEnableVertexAttribArray(material.getAttribLocation(attrNormal));		        
			normalBuffer.position(0);
		}
		
		if( colourBuffer != null ){
			GLES20.glVertexAttribPointer(material.getAttribLocation(attrColour), COLOUR_COMPONENT_COUNT, GL_FLOAT, false, 0, colourBuffer);
			GLES20.glEnableVertexAttribArray(material.getAttribLocation(attrColour));		        
			colourBuffer.position(0);
		}
		
		if( textureCoordsBuffer != null ){
			GLES20.glVertexAttribPointer(material.getAttribLocation(attrTextureCoords), TEXTURE_COMPONENT_COUNT, GL_FLOAT, false, 0, textureCoordsBuffer);
			GLES20.glEnableVertexAttribArray(material.getAttribLocation(attrTextureCoords));		        
			textureCoordsBuffer.position(0);
		}
				
	}
	
	@Override
	public void draw(){		
		GLES20.glDrawArrays(GLES20.GL_TRIANGLES, 0, vertexCount);
	}
	
	@Override
	public void unbind(Material material){
		GLES20.glDisableVertexAttribArray(material.getAttribLocation(Constants.ATTRIBUTE_POSITION));
		GLES20.glDisableVertexAttribArray(material.getAttribLocation(Constants.ATTRIBUTE_NORMAL));
		GLES20.glDisableVertexAttribArray(material.getAttribLocation(Constants.ATTRIBUTE_COLOUR));
		GLES20.glDisableVertexAttribArray(material.getAttribLocation(Constants.ATTRIBUTE_TEX_COORD));
	}
	
	@Override
	public void dispose(){
		
	}
}
