package com.wmp.nofrills3d.ik;

import com.wmp.nofrills3d.entities.SkinnedModel;
import com.wmp.nofrills3d.geometry.Bone;
import com.wmp.nofrills3d.math.MathHelper;
import com.wmp.nofrills3d.math.Matrix4;
import com.wmp.nofrills3d.math.Vector3;

public class IKChain extends IKConstraint{

	private float rotationChangeThreshold = 0.20943951f;
	
	public IKChain(SkinnedModel model, Bone bone, int length){
		super(model, bone, length);		 
	}
	
	/**
	 * CREDIT: http://xbox.create.msdn.com/en-US/education/catalog/sample/inverse_kinematics
	 * Loop over the IK chain and update each bone to bring the end effector closer to the target 
	 * @param elapsedTime
	 */
	@Override
	protected boolean solveIK(float elapsedTime){
				
		// iterating the bones reverse
        int index = 1;
        
        Vector3 goalInBoneWorldSpace = new Vector3(); 
        Matrix4 curBoneWorldInv = new Matrix4();
        Matrix4 endEffectorInBoneWorldSpaceMat = new Matrix4(); 
        Vector3 endEffectorInBoneWorldSpace = new Vector3(); 
        Matrix4 rotationAdjustment = new Matrix4(); 
        
        int updatedBoneCount = 0; 
        
        while (index <= chainLength)
        {
        	/**
        	 * 1) Compute vector directions for the bone to the goal and the bone to end effector.
        	 * 2) Compute a matrix that rotates the end effector vector onto the goal vector.
        	 * 3) Rotate the current bone by this rotation matrix.
        	 * 4) Repeat steps 1-3 for each bone in the chain.
        	 */
        	
        	Bone bone = ikBone.getParent(index);
        	
        	Matrix4 curBoneWorld = skinnedModel.worldTransforms[bone.index];        	        
        	
        	// Transform the goal into coordinate system of current bone. To do this, we
            // transform the goal position by the inverse world transform of the 
            // current bone.            
            Vector3.transform( target, Matrix4.invert(curBoneWorld, curBoneWorldInv), goalInBoneWorldSpace);
        	
            // Transform the end effector into coordinate system of the current bone. To 
            // do this, we first compute the current world transform of the end effector 
            // then we multiply it by the Inverse of the current bone's world transform. 
            // We then store the position.
            Matrix4 endEffectorWorld = skinnedModel.worldTransforms[ikBone.index];
            //Matrix4.decomposeTranslation(Matrix4.multiply(endEffectorWorld, curBoneWorldInv, endEffectorInBoneSpaceMat), endEffectorInBoneSpace);
            Matrix4.decomposeTranslation(Matrix4.multiply(curBoneWorldInv, endEffectorWorld, endEffectorInBoneWorldSpaceMat), endEffectorInBoneWorldSpace);
            
            // After we normalize, we will have unit vectors that represent the 
            // direction to the end effector and the goal in the local coordinate space 
            // of the current bone.
            endEffectorInBoneWorldSpace.normalise(); 
            goalInBoneWorldSpace.normalise();

            // Next we build the rotation matrix that rotates the end effector onto the
            // goal and apply that rotation to the current bone.

            // Compute axis of rotation: the cross product of the two vectors.
            Vector3 axis = Vector3.crossProduct(endEffectorInBoneWorldSpace, goalInBoneWorldSpace);

            // Use TransformNormal to orient the axis by the local coordinate
            // transform of the current bone
            //axis = Vector3.TransformNormal(axis, skinnedModel.boneTransforms[bone.index], axis);
            axis = Vector3.TransformNormal(axis, skinnedModel.boneTransforms[bone.index], axis);
            axis.normalise(); 
            
            // Compute the angle we will be rotating by which is just the angle 
            // between the vectors
            float dot = Vector3.dotProduct(goalInBoneWorldSpace, endEffectorInBoneWorldSpace);

            //Clamp to -1 and 1 to avoid any possible floating point precision errors
            dot = MathHelper.clamp(dot, -1, 1); 

            //Compute the angle.
            float angle = (float)Math.acos(dot); 
            angle = MathHelper.WrapAngle(angle);
            
            if( angle > rotationChangeThreshold ){
            	
            	updatedBoneCount++; 
            	// We can clamp the angle here which will make the animation look smoother. 
                // However, for demonstration purposes we will comment this out for now.
                //float clampAmount = MathHelper.deg2Rad(.5f);
                //angle = MathHelper.clamp(angle, -clampAmount, clampAmount);
                
                // Create the rotation matrix
                Matrix4.createFromAxisAngle(axis, angle, rotationAdjustment);            

                // Rotate the current bone by the new rotation matri            
                Matrix4 tmp = Matrix4.createIdentity(); 
                Matrix4.multiply(skinnedModel.boneTransforms[bone.index], rotationAdjustment, tmp);
                //Matrix4.multiply(rotationAdjustment, skinnedModel.boneTransforms[bone.index], tmp);
                //skinnedModel.boneTransforms[bone.index].set(tmp.matrix);
                Matrix4.blend(skinnedModel.boneTransforms[bone.index], tmp, ikSpeed, skinnedModel.boneTransforms[bone.index]);
            }                                    
        	
        	index++; 
        }
        
        return updatedBoneCount == 0 ? false : true; 
	}
}
