/**
 * Material.java (ColourExplosion)
 * Created by Joshua Newnham on Apr 9, 2013.
 * Copyright (c) 2013 We Make Play All rights reserved.
 */
package com.wmp.nofrills3d.material;

import static android.opengl.GLES20.GL_TEXTURE0;
import static android.opengl.GLES20.GL_TEXTURE_2D;
import static android.opengl.GLES20.glUniformMatrix4fv;
import android.content.Context;
import android.opengl.GLES20;

import com.wmp.nofrills3d.math.Matrix4;
import com.wmp.nofrills3d.scene.Constants;

public class Material extends Shader{

	public static final String TAG = "Material";		
	
	public Matrix4 projectionTransform = new Matrix4(); 
	
	public Matrix4 viewTransform = new Matrix4();
	
	public Matrix4 modelTransform = new Matrix4(); 
	
	public Texture texture; 		
	
	public Colour ambient = Colour.create(1f, 1f, 1f, 1f);
	
	public Colour diffuse = Colour.create(0.2f, 0.2f, 0.2f, 1.0f);
	
	public Colour specular = Colour.create(0.01f, 0.01f, 0.01f, 1.0f);
	
	public float shinness = 0.0f; 		
	
	public Light light = new Light(); 
	
	public int pass = 0; 
	
	protected Matrix4 modelViewTransform = new Matrix4(); 
	
	protected Matrix4 viewProjectionTransform = new Matrix4();		
	
	protected Matrix4 modelViewProjectionTransform = new Matrix4();
	
	protected Matrix4 normalMatrix = new Matrix4(); 		
	
	public Material(Context context, int vertexShaderResId, int fragmentShaderResId){
		super(context, vertexShaderResId, fragmentShaderResId);
	}		
	
	@Override
	public int bind(){
		/*
		 * order: 
			vertexndc = PerspectiveDivide(ProjectionMatrix * vertexeye)
			vertexndc = PerspectiveDivide(ProjectionMatrix * ViewMatrix * vertexworld)
			vertexndc = PerspectiveDivide(ProjectionMatrix * ViewMatrix * ModelMatrix * vertexmodel)
		 */
		// create modelViewTransform from model and transform transformations 
		
		Matrix4.multiply(viewTransform, modelTransform, modelViewTransform);
		//Matrix4.multiply(projectionTransform, modelViewTransform, modelViewProjectionTransform);		
		
		Matrix4.multiply(projectionTransform, viewTransform, viewProjectionTransform);
		Matrix4.multiply(viewProjectionTransform, modelTransform, modelViewProjectionTransform);				
		
		// Create the normal modelview matrix (Invert + transpose of model view projection matrix) 		
		Matrix4.invert(modelViewProjectionTransform, normalMatrix);
		Matrix4.transpose(normalMatrix, normalMatrix);
		
		super.bind(); 
		
		// set uniforms
		int uMVPtr = getUniformLocation(Constants.UNIFORM_MV); 
		if( uMVPtr != -1 ){
			glUniformMatrix4fv(uMVPtr, 1, false, modelViewTransform.matrix, 0);
		}
				
		int uMVPPtr = getUniformLocation(Constants.UNIFORM_MVP); 
		if( uMVPPtr != -1 ){
			glUniformMatrix4fv(uMVPPtr, 1, false, modelViewProjectionTransform.matrix, 0);
		}
		
		int uNormalPtr = getUniformLocation(Constants.UNIFORM_NORMAL);
		if( uNormalPtr != -1 ){
			glUniformMatrix4fv(uNormalPtr, 1, false, normalMatrix.matrix, 0);
		}
		
		if( texture != null ){
			int uTexPtr = getUniformLocation(Constants.UNIFORM_TEXTURE_0);				        
	        if( uTexPtr != -1 ){
	        	GLES20.glActiveTexture(GL_TEXTURE0);	        
		        GLES20.glBindTexture(GL_TEXTURE_2D, texture.textureId);
	        	GLES20.glUniform1i(uTexPtr, 0);
	        }
	        
	        int uTexAvailablePtr = getUniformLocation(Constants.UNIFORM_TEXTURE_AVAILABLE);				        
	        if( uTexAvailablePtr != -1 ){
	        	GLES20.glUniform1i(uTexAvailablePtr, 1);
	        }
		} else{
			int uTexAvailablePtr = getUniformLocation(Constants.UNIFORM_TEXTURE_AVAILABLE);				        
	        if( uTexAvailablePtr != -1 ){
	        	GLES20.glUniform1i(uTexAvailablePtr, -1);
	        }
		}				
		
		int uAmbientPtr = getUniformLocation(Constants.UNIFORM_AMBIENT);
		if( uAmbientPtr != -1 ){
			GLES20.glUniform4f(uAmbientPtr, ambient.colour[Colour.R], ambient.colour[Colour.G], ambient.colour[Colour.B], ambient.colour[Colour.A]);
		}
		
		int uDiffusePtr = getUniformLocation(Constants.UNIFORM_DIFFUSE);
		if( uDiffusePtr != -1 ){
			GLES20.glUniform4f(uDiffusePtr, diffuse.colour[Colour.R], diffuse.colour[Colour.G], diffuse.colour[Colour.B], diffuse.colour[Colour.A]);
		}
		
		int uSpecularPtr = getUniformLocation(Constants.UNIFORM_SPECULAR);
		if( uSpecularPtr != -1 ){
			GLES20.glUniform4f(uSpecularPtr, specular.colour[Colour.R], specular.colour[Colour.G], specular.colour[Colour.B], specular.colour[Colour.A]);
		}
		
		int uShinnessPtr = getUniformLocation(Constants.UNIFORM_SHINNESS);
		if( uShinnessPtr != -1 ){
			GLES20.glUniform1f(uShinnessPtr, shinness);
		}
		
		bindLight();				
		
		return program;
	}
	
	protected void bindLight(){
		int uLightOnPtr = getUniformLocation(Constants.UNIFORM_LIGHT_ENABLED);
		if( uLightOnPtr != -1 ){
			GLES20.glUniform1i(uLightOnPtr, (light.enabled ? 1 : 0)); 
		}
		
		int uLightPositionPtr = getUniformLocation(Constants.UNIFORM_LIGHT_POSITION);
		if( uLightPositionPtr != -1 ){
			//GLES20.glUniform3f(uLightPositionPtr, light.lightPosition.x, light.lightPosition.y, light.lightPosition.z);
			GLES20.glUniform4fv(uLightPositionPtr, 1, light.lightPosition, 0);
		}
		
		int uLightDiffusePtr = getUniformLocation(Constants.UNIFORM_LIGHT_DIFFUSE);
		if( uLightDiffusePtr != -1 ){
			GLES20.glUniform4f(uLightDiffusePtr, light.diffuse.colour[Colour.R], light.diffuse.colour[Colour.G], light.diffuse.colour[Colour.B], light.diffuse.colour[Colour.A]);
		}
		
		int uLightAmbientPtr = getUniformLocation(Constants.UNIFORM_LIGHT_AMBIENT);
		if( uLightAmbientPtr != -1 ){
			GLES20.glUniform4f(uLightAmbientPtr, light.ambient.colour[Colour.R], light.ambient.colour[Colour.G], light.ambient.colour[Colour.B], light.ambient.colour[Colour.A]);
		}
		
		int uLightSpecularPtr = getUniformLocation(Constants.UNIFORM_LIGHT_SPECULAR);
		if( uLightSpecularPtr != -1 ){
			GLES20.glUniform4f(uLightSpecularPtr, light.specular.colour[Colour.R], light.specular.colour[Colour.G], light.specular.colour[Colour.B], light.diffuse.colour[Colour.A]);
		}
	}
	
	@Override
	public void unbind(){
		if( texture != null ){	        
	        GLES20.glBindTexture(GL_TEXTURE_2D, 0);
		}
		
		super.unbind(); 
	}
	
	@Override 
	public void dispose(){
		super.dispose(); 
	}		
}
