/**
 * NoFrills3D
 * com.wmp.nofrills3d.material
 * 
 * @author Joshua Newnham
 * @copyright We Make Play Ltd. 2013 
 * 
 */
package com.wmp.nofrills3d.material;

import static android.opengl.GLES20.GL_COMPILE_STATUS;
import static android.opengl.GLES20.GL_FRAGMENT_SHADER;
import static android.opengl.GLES20.GL_LINK_STATUS;
import static android.opengl.GLES20.GL_VALIDATE_STATUS;
import static android.opengl.GLES20.GL_VERTEX_SHADER;
import static android.opengl.GLES20.glAttachShader;
import static android.opengl.GLES20.glCompileShader;
import static android.opengl.GLES20.glCreateProgram;
import static android.opengl.GLES20.glCreateShader;
import static android.opengl.GLES20.glDeleteProgram;
import static android.opengl.GLES20.glDeleteShader;
import static android.opengl.GLES20.glGetProgramInfoLog;
import static android.opengl.GLES20.glGetProgramiv;
import static android.opengl.GLES20.glGetShaderiv;
import static android.opengl.GLES20.glLinkProgram;
import static android.opengl.GLES20.glShaderSource;
import static android.opengl.GLES20.glValidateProgram;
import android.content.Context;
import android.opengl.GLES20;

import com.wmp.nofrills3d.utils.ContentLoader;
import com.wmp.nofrills3d.utils.Logger;

/**
 * 
 */
public class Shader {
	
	public static final String TAG = "Shader";

	protected int program;
	
	protected int vertexShaderId; 
	
	protected int fragmentShaderId;
	
	public Shader(Context context, int vertexShaderResId, int fragmentShaderResId){
		String vertexShader = ContentLoader.loadTextFileFromResource(context, vertexShaderResId);
		String fragmentShader = ContentLoader.loadTextFileFromResource(context, fragmentShaderResId);
		
		buildShader(vertexShader, fragmentShader);
	}
	
	public int getProgram(){
		return program; 
	}
	
	public int bind(){
		GLES20.glUseProgram(program);
		checkGlError("glUseProgram");
		
		return program; 
	}
	
	public void unbind(){
		GLES20.glUseProgram(0);		
	}
	
	public void dispose(){
		if( program == -1 ){
			return; 
		}
		
		GLES20.glDeleteShader(vertexShaderId);
		GLES20.glDeleteShader(fragmentShaderId);
		GLES20.glDeleteProgram(program);
		
		program = -1; 
	}
	
	public int getAttribLocation(String attribute){
		return GLES20.glGetAttribLocation(program, attribute);
	}
	
	public int getUniformLocation(String unifrom){
		return GLES20.glGetUniformLocation(program, unifrom);
	}	
	
	protected void checkGlError(String op) {
		int error;
		while ((error = GLES20.glGetError()) != GLES20.GL_NO_ERROR) {
			Logger.LogError(TAG, op + ": glError " + error);
			throw new RuntimeException(op + ": glError " + error);
		}
	}
	
	protected int buildShader(String vertexShaderSource, String fragmentShaderSource){

        vertexShaderId = compileVertexShader(vertexShaderSource);
        
        Logger.Log(TAG, (vertexShaderId > 0 ? "Loaded: VS \n" : "Failed to load VS: \n") + vertexShaderSource);
        
        fragmentShaderId = compileFragmentShader(fragmentShaderSource);
        
        Logger.Log(TAG, (fragmentShaderId > 0 ? "Loaded FS: \n" : "Failed to load FS: \n") + fragmentShaderSource);

        program = linkProgram(vertexShaderId, fragmentShaderId);

        if (Logger.Logging) {
            validateProgram(program);
        }

        return program;
	}
	
	public int compileVertexShader(String shaderCode) {
        return compileShader(GL_VERTEX_SHADER, shaderCode);
    }

    /**
     * Loads and compiles a fragment shader, returning the OpenGL object ID.
     */
    public int compileFragmentShader(String shaderCode) {
        return compileShader(GL_FRAGMENT_SHADER, shaderCode);
    }

    /**
     * Compiles a shader, returning the OpenGL object ID.
     */
    private int compileShader(int type, String shaderCode) {
        final int shaderObjectId = glCreateShader(type);

        if (shaderObjectId == 0) {
            Logger.LogError(TAG, "Could not create new shader.");
            return 0;
        }

        glShaderSource(shaderObjectId, shaderCode);
        glCompileShader(shaderObjectId);

        final int[] compileStatus = new int[1];
        glGetShaderiv(shaderObjectId, GL_COMPILE_STATUS, compileStatus, 0);

        if (compileStatus[0] == 0) {
            glDeleteShader(shaderObjectId);

            Logger.LogError(TAG, "Compilation of shader failed.");

            return 0;
        }

        return shaderObjectId;
    }

    /**
     * Links a vertex shader and a fragment shader together into an OpenGL
     * program. Returns the OpenGL program object ID, or 0 if linking failed.
     */
    public int linkProgram(int vertexShaderId, int fragmentShaderId) {

        final int programObjectId = glCreateProgram();

        if (programObjectId == 0) {
        	Logger.LogError(TAG, "Could not create new program");
            return 0;
        }

        // Attach the vertex shader to the program.
        glAttachShader(programObjectId, vertexShaderId);

        // Attach the fragment shader to the program.
        glAttachShader(programObjectId, fragmentShaderId);

        // Link the two shaders together into a program.
        glLinkProgram(programObjectId);

        // Get the link status.
        final int[] linkStatus = new int[1];
        glGetProgramiv(programObjectId, GL_LINK_STATUS,
            linkStatus, 0);

        // Verify the link status.
        if (linkStatus[0] == 0) {
            // If it failed, delete the program object.
            glDeleteProgram(programObjectId);
            Logger.LogWarning(TAG, "Linking of program failed.");
            return 0;
        }

        // Return the program object ID.
        return programObjectId;
    }

    /**
     * Validates an OpenGL program. Should only be called when developing the
     * application.
     */
    public static boolean validateProgram(int programObjectId) {
        glValidateProgram(programObjectId);
        final int[] validateStatus = new int[1];
        glGetProgramiv(programObjectId, GL_VALIDATE_STATUS, validateStatus, 0);
        
        Logger.Log(TAG, "Results of validating program: " + validateStatus[0] + "\nLog:" + glGetProgramInfoLog(programObjectId));

        return validateStatus[0] != 0;
    }
	
}
