/**
 * MathHelper.java (ColourExplosion)
 * Created by Joshua Newnham on Apr 9, 2013.
 * Copyright (c) 2013 We Make Play All rights reserved.
 */
package com.wmp.nofrills3d.math;

public final class MathHelper {

	public static final int X = 0; 
	public static final int Y = 1;
	public static final int Z = 2;
	public static final int W = 3;
	
	public static final float PI_OVER_2 = (float)(Math.PI / 2.0f);
	
	public static final float PI_OVER_4 = (float)(Math.PI / 4.0f);
	
	public static final float TWO_PI = (float)(Math.PI * 2.0f);		
	
	public static float rad2Deg(float rad){
		return (float)(rad * 180.0f / Math.PI);
	}
	
	public static float deg2Rad(float deg){
		return (float) (deg * Math.PI / 180.0); 
	}
	
	public static void rotateAroundPoint(float[] position, float[] quaternion, float[] result){
		float x = 2 * (quaternion[Y] * position[Z] - quaternion[Z] * position[Y]);
        float y = 2 * (quaternion[Z] * position[X] - quaternion[X] * position[Z]);
        float z = 2 * (quaternion[X] * position[Y] - quaternion[Y] * position[X]);

        result[X] = position[X] + x * quaternion[W] + (quaternion[Y] * z - quaternion[Z] * y);
        result[Y] = position[Y] + y * quaternion[W] + (quaternion[Z] * x - quaternion[X] * z);
        result[Z] = position[Z] + z * quaternion[W] + (quaternion[X] * y - quaternion[Y] * x);
	}	
	
	public static float clamp( float val, float min, float max ){
		return Math.max(Math.min(val, max), min);
	}
	
	public static float WrapAngle(float angle){
        angle = (float)Math.IEEEremainder((double)angle, 6.2831854820251465);

		if (angle <= -3.14159274f) {
			angle += 6.28318548f;
		} else {
			if (angle > 3.14159274f) {
				angle -= 6.28318548f;
			}
		}
		return angle;
	}
	
	/**
	 * Performs a Catmull-Rom interpolation using the specified positions.
	 * @param value1 The first position in the interpolation
	 * @param value2 The second position in the interpolation.
	 * @param value3 The third position in the interpolation.
	 * @param value4 The forth position in the interpolation.
	 * @param amount weighting factor 
	 * @return A position that is the result of the Catmull-Rom interpolation.
	 */
	public static float catmullRom(float value1, float value2, float value3, float value4, float amount){
        // Using formula from http://www.mvps.org/directx/articles/catmull/
        // Internally using doubles not to lose precission
        double amountSquared = amount * amount;
        double amountCubed = amountSquared * amount;
        return (float)(0.5 * (2.0 * value2 +
            (value3 - value1) * amount +
            (2.0 * value1 - 5.0 * value2 + 4.0 * value3 - value4) * amountSquared +
            (3.0 * value2 - value1 - 3.0 * value3 + value4) * amountCubed));
    }
	
	/**
	 * Returns the Cartesian coordinate for one axis of a point that is defined by a given triangle and two normalized barycentric (areal) coordinates.
	 * @param value1 The coordinate on one axis of vertex 1 of the defining triangle
	 * @param value2 The coordinate on the same axis of vertex 2 of the defining triangle
	 * @param value3 The coordinate on the same axis of vertex 3 of the defining triangle
	 * @param amount1 The normalized barycentric (areal) coordinate b2, equal to the weighting factor for vertex 2, the coordinate of which is specified in value2
	 * @param amount2 The normalized barycentric (areal) coordinate b3, equal to the weighting factor for vertex 3, the coordinate of which is specified in value3
	 * @return Cartesian coordinate of the specified point with respect to the axis being used
	 */
	public static float Barycentric(float value1, float value2, float value3, float amount1, float amount2){
        return value1 + (value2 - value1) * amount1 + (value3 - value1) * amount2;
    }
	
	/**
	 * 
	 * @param value1
	 * @param value2
	 * @return
	 */
	public static float distance(float value1, float value2){
        return Math.abs(value1 - value2);
    }
	
	/**
	 * Hermite spline interpolation
	 * @param value1
	 * @param tangent1
	 * @param value2
	 * @param tangent2
	 * @param amount
	 * @return
	 */
	public static float hermite(float value1, float tangent1, float value2, float tangent2, float amount)
    {
        // All transformed to double not to lose precission
        // Otherwise, for high numbers of param:amount the result is NaN instead of Infinity
        double v1 = value1, v2 = value2, t1 = tangent1, t2 = tangent2, s = amount, result;
        double sCubed = s * s * s;
        double sSquared = s * s;

        if (amount == 0f)
            result = value1;
        else if (amount == 1f)
            result = value2;
        else
            result = (2 * v1 - 2 * v2 + t2 + t1) * sCubed +
                (3 * v2 - 3 * v1 - 2 * t1 - t2) * sSquared +
                t1 * s +
                v1;
        return (float)result;
    }
	
	/**
	 * linear interpolation between 2 values 
	 * @param value1
	 * @param value2
	 * @param amount
	 * @return
	 */
	public static float lerp(float value1, float value2, float amount){
        return value1 + (value2 - value1) * amount;
    }
	
	/**
	 * returns an angle between 2 vectors 
	 * @param v1
	 * @param v2
	 * @return
	 */
	public static float angleBetween(Vector3 v1, Vector3 v2){
		// See http://en.wikipedia.org/wiki/Vector_(spatial)
        // for help and check out the Dot Product section ^^
        // Both vectors are normalized so we can save deviding through the
        // lengths.
		float dot = Vector3.dotProduct(v1, v2);
		return (float)Math.acos(dot);
	}		
}
