/**
 * Vector3.java (ColourExplosion)
 * Created by Joshua Newnham on Apr 9, 2013.
 * Copyright (c) 2013 We Make Play All rights reserved.
 */
package com.wmp.nofrills3d.math;

import android.annotation.SuppressLint;




public class Vector3 {

	public static final Vector3 UP = new Vector3(0,1,0);
	
	public static final Vector3 RIGHT = new Vector3(1,0,0);
	
	public static final Vector3 FORWARD = new Vector3(0,0,1);
	
	public static Vector3 createZero(){ return new Vector3(); }	
	
	public float x = 0; 
	public float y = 0; 
	public float z = 0; 
	
	protected float[] vector = new float[3]; 
	
	public Vector3(){
		this.x = 0; this.y = 0; this.z = 0;  
	}
	
	public Vector3(float val){
		this.x = val; this.y = val; this.z = val;
	}
	
	public Vector3(Vector3 val){
		this.x = val.x; this.y = val.y; this.z = val.z;
	}
	
	public Vector3(float x, float y, float z){
		this.x = x; this.y = y; this.z = z; 
	}		
	
	public Vector3 set(float x, float y, float z){
		this.x = x; 
		this.y = y; 
		this.z = z; 
		
		return this; 
	}
	
	public Vector3 set(Vector3 other){
		this.x = other.x; 
		this.y = other.y; 
		this.z = other.z; 
		
		return this; 
	}
	
	public Vector3 set(float[] v){
		this.x = v[0]; 
		this.y = v[1]; 
		this.z = v[2];
		
		return this; 
	}
	
	public Vector3 normalise(){
		float len = length(); 
		this.x /= len; 
		this.y /= len; 
		this.z /= len; 
		
		return this; 
	}
	
	public float length(){
		return (float)Math.sqrt(x * x + y * y + z * z);		
	}
	
	public float lengthSqr(){
		return x * x + y * y + z * z;		
	}
	
	public Vector3 scale(float scale){
		this.x *= scale;
		this.y *= scale; 
		this.z *= scale; 
		
		return this; 
	}
	
	public Vector3 scale(Vector3 scale){
		this.x *= scale.x;
		this.y *= scale.y; 
		this.z *= scale.z; 
		
		return this; 
	}
	
	public Vector3 add(Vector3 other){
		this.x += other.x; 
		this.y += other.y; 
		this.z += other.z;
		
		return this; 
	}
	
	public float[] getFloatArray(){
		vector[0] = x; vector[1] = y; vector[2] = z; 		
		return vector; 
	}
	
	public Vector3 clone(){
		return new Vector3(this.x, this.y, this.z);
	}
	
	@SuppressLint("DefaultLocale")
	@Override
	public String toString(){
		return String.format("V(%f,%f,%f)", x, y, z);
	}
	
	public static Vector3 createOne(){ return new Vector3(1,1,1); }
	
	public static Vector3 subtract( Vector3 v1, Vector3 v2 ){
		return new Vector3(v2.x - v1.x, v2.y - v1.y, v2.z - v1.z);
	}
	
	public static void subtract( Vector3 v1, Vector3 v2, Vector3 result ){
		result.set(v2.x - v1.x, v2.y - v1.y, v2.z - v1.z);
	}
	
	public static Vector3 add( Vector3 v1, Vector3 v2 ){
		return new Vector3(v1.x + v2.x, v1.y + v2.y, v1.z + v2.z);
	}
	
	public static void add( Vector3 v1, Vector3 v2, Vector3 result ){
		result.set(v1.x + v2.x, v1.y + v2.y, v1.z + v2.z);
	}
	
	public static Vector3 scale( Vector3 v1, float scale ){
		return new Vector3(v1.x * scale, v1.y * scale, v1.z * scale);
	}
	
	public static void scale( Vector3 v1, float scale, Vector3 result ){
		result.set(v1.x * scale, v1.y * scale, v1.z * scale);
	}
	
	public static void direction(Vector3 from, Vector3 to, Vector3 result) {
		result.set(
            to.x - from.x, 
            to.y - from.y, 
            to.z - from.z);
		
		result.normalise(); 
    }
	
	public static float distanceSqr(Vector3 vector1, Vector3 vector2) {
		float dx = vector2.x - vector1.x;
		float dy = vector2.y - vector1.y;
		float dz = vector2.z - vector1.z;
		
		return dx * dx + dy * dy + dz * dz; 
    }
	
	public static float distance(Vector3 vector1, Vector3 vector2) {
		float distanceSqr = distanceSqr(vector1, vector2);
		return (float)Math.sqrt(distanceSqr);
    }
	
	/**
	 * results in a vector which is perpendicular to both of the vectors being multiplied and therefore normal to the plane containing them
	 * @param vector1
	 * @param vector2
	 * @param result
	 */
    public static void crossProduct(Vector3 vector1, Vector3 vector2, Vector3 result) {
    	result.set((vector1.y * vector2.z) - (vector1.z * vector2.y), 
                (vector1.z * vector2.x) - (vector1.x * vector2.z), 
                (vector1.x * vector2.y) - (vector1.y * vector2.x));
    }
    
    public static Vector3 crossProduct(Vector3 vector1, Vector3 vector2) {        
    	return new Vector3((vector1.y * vector2.z) - (vector1.z * vector2.y), 
                (vector1.z * vector2.x) - (vector1.x * vector2.z), 
                (vector1.x * vector2.y) - (vector1.y * vector2.x));
    }    

    /**
     * dot product is directly related to the cosine of the angle between two vectors in Euclidean space
     * @param vector1
     * @param vector2
     * @return
     */
    public static float dotProduct(Vector3 vector1, Vector3 vector2) {
        return vector1.x * vector2.x 
             + vector1.y * vector2.y 
             + vector1.z * vector2.z;
    }
    
    public static Vector3 min(Vector3 value1, Vector3 value2)
    {
        return new Vector3(
            (float)Math.min(value1.x, value2.x),
            (float)Math.min(value1.y, value2.y),
            (float)Math.min(value1.z, value2.z));
    }
	
	public static Vector3 max(Vector3 value1, Vector3 value2)
    {
        return new Vector3(
            (float)Math.max(value1.x, value2.x),
            (float)Math.max(value1.y, value2.y),
            (float)Math.max(value1.z, value2.z));
    }
	
	public static Vector3 project(Vector3 vec, Matrix4 mat){
		return project(vec, mat, new Vector3());
	}
	
	public static Vector3 project(Vector3 vec, Matrix4 mat, Vector3 res){
		
		float x = vec.x; 
		float y = vec.y; 
		float z = vec.z; 
		
        float l_w = x * mat.matrix[Matrix4.M14] + y * mat.matrix[Matrix4.M24] + z * mat.matrix[Matrix4.M34] + mat.matrix[Matrix4.M44];
        
        res.set(
      		  (x * mat.matrix[Matrix4.M11] + y * mat.matrix[Matrix4.M21] + z * mat.matrix[Matrix4.M31] + mat.matrix[Matrix4.M41]) / l_w, 
      		  (x * mat.matrix[Matrix4.M12] + y * mat.matrix[Matrix4.M22] + z * mat.matrix[Matrix4.M32] + mat.matrix[Matrix4.M42]) / l_w, 
      		  (x * mat.matrix[Matrix4.M13] + y * mat.matrix[Matrix4.M23] + z * mat.matrix[Matrix4.M33] + mat.matrix[Matrix4.M43]) / l_w);
        
        return res;         
	}
	
	public static Vector3 transform(Vector3 vec, Matrix4 mat)
    {
		Vector3 result = new Vector3(); 
		return transform(vec, mat, result);
    }
	
	public static Vector3 transform(Vector3 vec, Matrix4 mat, Vector3 result)
    {
        result.set(
        		(vec.x * mat.matrix[Matrix4.M11]) + (vec.y * mat.matrix[Matrix4.M21]) + (vec.z * mat.matrix[Matrix4.M31]) + mat.matrix[Matrix4.M41],
        		(vec.x * mat.matrix[Matrix4.M12]) + (vec.y * mat.matrix[Matrix4.M22]) + (vec.z * mat.matrix[Matrix4.M32]) + mat.matrix[Matrix4.M42],
                (vec.x * mat.matrix[Matrix4.M13]) + (vec.y * mat.matrix[Matrix4.M23]) + (vec.z * mat.matrix[Matrix4.M33]) + mat.matrix[Matrix4.M43]);
        
        return result; 
    }
	
	public static Vector3 TransformNormal(Vector3 normal, Matrix4 mat, Vector3 result)
    {
        result.set(
        		(normal.x * mat.matrix[Matrix4.M11]) + (normal.y * mat.matrix[Matrix4.M21]) + (normal.z * mat.matrix[Matrix4.M31]),
                (normal.x * mat.matrix[Matrix4.M12]) + (normal.y * mat.matrix[Matrix4.M22]) + (normal.z * mat.matrix[Matrix4.M32]),
                (normal.x * mat.matrix[Matrix4.M13]) + (normal.y * mat.matrix[Matrix4.M23]) + (normal.z * mat.matrix[Matrix4.M33]));
        
        return result; 
    }
	
	public static Vector3 lerp(Vector3 value1, Vector3 value2, float amount)
    {
		Vector3 result = new Vector3();
		return Vector3.lerp(value1, value2, amount, result);
    }

    public static Vector3 lerp(Vector3 value1, Vector3 value2, float amount, Vector3 result)
    {
        result.set(
            MathHelper.lerp(value1.x, value2.x, amount),
            MathHelper.lerp(value1.y, value2.y, amount),
            MathHelper.lerp(value1.z, value2.z, amount));
        
        return result; 
    }
}
