/**
 * NoFrills3D
 * com.wmp.nofrills3d.scene
 * 
 * @author Joshua Newnham
 * @copyright We Make Play Ltd. 2013 
 * 
 */
package com.wmp.nofrills3d.scene;

import static android.opengl.GLES20.glClear;

import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import android.content.Context;
import android.opengl.GLES20;

import com.wmp.colourexplosion.R;
import com.wmp.nofrills3d.effects.EffectsProcessor;
import com.wmp.nofrills3d.effects.RenderTarget;
import com.wmp.nofrills3d.effects.TextureRenderer;
import com.wmp.nofrills3d.entities.Camera;
import com.wmp.nofrills3d.entities.Entity;
import com.wmp.nofrills3d.entities.Model;
import com.wmp.nofrills3d.input.UserInputEvent;
import com.wmp.nofrills3d.material.Shader;
import com.wmp.nofrills3d.material.ShadowMaterial;
import com.wmp.nofrills3d.math.Ray;
import com.wmp.nofrills3d.physics.ContactCollection;

/**
 * 
 */
public class Scene implements IScene {		
	
	protected SceneRenderer sceneRenderer = null; 
	
	protected Camera camera; 	
	
	protected List<Entity> entities = new Vector<Entity>();
	
	protected boolean shadowsEnabled = false; 
	
	protected ContactCollection collisions = new ContactCollection();	
	
	protected RenderTarget shadowRenderTarget = null; 
	
	protected RenderTarget offscreenRenderTarget = null;
	
	protected TextureRenderer offscreenTextureRenderer = null;
	
	protected Shader offscreenShader = null; 
	
	protected EffectsProcessor effectsProcessor = null; 
	
	public Scene(){
		
	}
	
	public void setShadowsEnabled(boolean shadowsEnabled){
		this.shadowsEnabled = shadowsEnabled; 
	}
	
	public boolean isShadowsEnabled(){
		return this.shadowsEnabled; 
	}
	
	public RenderTarget getShadowRenderTarget(){
		return shadowRenderTarget;
	}
	
	public void setSceneRenderer(SceneRenderer renderer){
		this.sceneRenderer = renderer; 
	}
	
	public SceneRenderer getSceneRenderer(){
		return this.sceneRenderer; 
	}
	
	public void setCamera(Camera camera){
		this.camera = camera; 
	}	
	
	public Camera getCamera(){
		return this.camera; 
	}

	public Iterator<Entity> getEntityIterator(){
		return entities.iterator(); 
	}
	
	public void addEntity(Entity entity){
		entities.add(entity);
		
		if( !entity.isLoaded() ){
			entity.load(this);
		}				
	}
	
	public void removeEntity(Entity entity){
		entities.remove(entity);
		
		if( entity.isLoaded() ){
			entity.unload(true);
		}		
	}
	
	public Entity findEntityWithTag(String tag){
		for( Entity entity : entities ){
			if( entity.tag.equals(tag) ){
				return entity; 
			}
		}
		
		return null; 
	}
	
	public Ray screenPositionToRay(float x, float y){
		return camera.screenPositionToRay(x, y);
	}
	
	public void onSceneRendererSurfaceChanged(SceneRenderer renderer, int viewWidth, int viewHeight){
		if( camera != null ){
			camera.viewWidth = viewWidth; 
			camera.viewHeight = viewHeight; 
			camera.init(); 
		}
		
		initOffscreenRenderer(); 
	}
	
	protected void initOffscreenRenderer(){
		
		if( shadowsEnabled ){
			if( shadowRenderTarget == null ){
				shadowRenderTarget = new RenderTarget(RenderTarget.RENDER_TO_TEXTURE | RenderTarget.RENDER_TO_DEPTH_RENDER_BUFFER, 
						(int)(1024), (int)(1024), null);		
			}
			
			// update material 
//			for( Entity entity : entities ){
//				if( entity instanceof Model ){
//					Model model = (Model)entity;
//					if( model.material instanceof ShadowMaterial ){
//						((ShadowMaterial)model.material).setShadowMapTexture(shadowRenderTarget.getTexture());
//					}
//				}								
//			}
		}
		
		if( offscreenRenderTarget == null ){
			offscreenRenderTarget = new RenderTarget(RenderTarget.RENDER_TO_TEXTURE | RenderTarget.RENDER_TO_DEPTH_RENDER_BUFFER, 
					getViewWidth(), getViewHeight(), null);		
		}
		
		if( offscreenTextureRenderer == null ){
			offscreenTextureRenderer = new TextureRenderer(getViewWidth(), getViewHeight()); 
		}
		
		if( offscreenShader == null ){
			offscreenShader = new Shader(getContext(), R.raw.texture_renderer_vs, R.raw.texture_renderer_fs);
		}
		
		if( effectsProcessor == null ){
			effectsProcessor = new EffectsProcessor(offscreenRenderTarget);
		}
	}
	
	public void onUpdate(float elapsedtime){
		for( Entity entity : entities ){
			if( entity.enabled ){
				entity.update(elapsedtime);
			}
		}
        
        for( Entity entity : entities ){
			if( entity.enabled ){
				entity.postUpdate(elapsedtime);
			}
		}
        
        camera.update(elapsedtime);
        
        collisionDetection();
	}
	
	protected void collisionDetection(){						
		
		int entityCount = entities.size(); 
		for( int i=0; i<entityCount; i++ ){
			Entity entity = entities.get(i);
			
			if( !entity.isCollidable() || !entity.enabled ){
				continue; 
			}
			
			// iterate through the other entities 
			for( int j=i+1; j<entityCount; j++ ){
				Entity otherEntity = entities.get(j);
				
				if( !otherEntity.isCollidable() || !otherEntity.enabled ){
					continue; 
				}								
				
				collisions.clear(); 								
												
				if( entity.collidesWith(otherEntity, collisions) > 0 ){
					entity.onHandleCollisions(collisions);
					otherEntity.onHandleCollisions(collisions);
				}
			}
		}		
	}
	
	public void onRender(){		
		
		// cull backface
		GLES20.glEnable( GLES20.GL_CULL_FACE );
		GLES20.glCullFace(GLES20.GL_BACK);
		
		if( shadowsEnabled ){
			GLES20.glViewport(0, 0, shadowRenderTarget.getTexture().width, shadowRenderTarget.getTexture().height);
			shadowRenderTarget.bind();
			
			renderSceneEntities(0, shadowsEnabled);
			
			shadowRenderTarget.unbind(); 
			
			GLES20.glViewport(0, 0, getViewWidth(), getViewHeight());
			
			renderSceneEntities(1, shadowsEnabled);
			//offscreenTextureRenderer.render(offscreenShader, shadowRenderTarget.getTexture());
		} 
		
		else{
			GLES20.glViewport(0, 0, getViewWidth(), getViewHeight());
			
			renderSceneEntities(0, shadowsEnabled);
		}
		
//		else if( effectsProcessor.hasEffects() ){
//			offscreenRenderTarget.bind();
//		}
//		
//		if( shadowsEnabled ){						
//			
//			
//		} else{
//			// cull backface
//			GLES20.glEnable( GLES20.GL_CULL_FACE );
//			GLES20.glCullFace(GLES20.GL_BACK);
//			
//			renderSceneEntities(1, shadowsEnabled);
//		}				
//		
//		if( effectsProcessor.hasEffects() ){
//			effectsProcessor.process(offscreenRenderTarget.getTexture());
//		}
		
//		if( shadowsEnabled || !effectsProcessor.hasEffects() ){
//			offscreenRenderTarget.unbind(); 
//		
//			offscreenTextureRenderer.render(offscreenShader, offscreenRenderTarget.getTexture());
//		}
	}
	
	protected void renderSceneEntities(int pass, boolean shadowsEnabled){
		GLES20.glEnable(GLES20.GL_DEPTH_TEST);					
		glClear(GLES20.GL_COLOR_BUFFER_BIT | GLES20.GL_DEPTH_BUFFER_BIT);					
		
		for( Entity entity : entities ){			
			
			if( entity.visible ){
				entity.render(camera.projectionMatrix, camera.modelTransform, pass);							
			}
		}
	}
	
	protected Context getContext(){
		if( sceneRenderer != null ){
			return sceneRenderer.getContext();
		}
		
		return null;
	}
	
	protected int getViewWidth(){
		if( sceneRenderer != null ){
			return sceneRenderer.viewWidth;
		}
		
		return 0;
	}
	
	protected int getViewHeight(){
		if( sceneRenderer != null ){
			return sceneRenderer.viewHeight;
		}
		
		return 0;
	}
	
	protected Iterator<UserInputEvent> getUserInputEventsIterator(){
		if( sceneRenderer != null ){
			return sceneRenderer.getUserInputEventsIterator();
		}
		
		return null;
	}
		
	public void flush(){
		for( int i=0; i<entities.size(); i++ ){
			this.entities.get(i).unload(false);
		}
		
		this.entities.clear(); 
		
		if( shadowRenderTarget != null ){
			shadowRenderTarget = null; 
		}
		
		if( offscreenRenderTarget != null ){
			offscreenRenderTarget = null; 
		}
		
		if( offscreenTextureRenderer != null ){
			offscreenTextureRenderer = null; 
		}
		
		if( offscreenShader != null ){
			offscreenShader = null; 
		}
		
		if( effectsProcessor != null ){
			effectsProcessor.dispose(); 
			effectsProcessor = null; 
		}
	}
}
