/**
 * GLRenderer.java (ColourExplosion)
 * Created by Joshua Newnham on Apr 9, 2013.
 * Copyright (c) 2013 We Make Play All rights reserved.
 */
package com.wmp.nofrills3d.scene;

import static android.opengl.GLES20.glClearColor;

import java.nio.IntBuffer;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import javax.microedition.khronos.egl.EGLConfig;
import javax.microedition.khronos.opengles.GL10;

import android.content.Context;
import android.graphics.Bitmap;
import android.opengl.GLES20;
import android.opengl.GLSurfaceView.Renderer;
import android.view.View;

import com.wmp.nofrills3d.input.UserInputEvent;
import com.wmp.nofrills3d.input.UserInputManager;

public class SceneRenderer implements Renderer{			 
	
	protected ISceneRendererDelegate delegate;
	
	protected int viewWidth;
	
	protected int viewHeight; 
	
	protected View view; 
	
	protected Context context; 		 
	
	protected long startTime = System.nanoTime();
	
	protected long frameRate = Constants.DEFAULT_FRAME_RATE; 
	
	protected long lastRenderTimestamp = System.currentTimeMillis();
	
	protected boolean depthTestEnabled = true; 		
	
	protected UserInputManager inputManager;
	
	protected List<UserInputEvent> inputEvents = new Vector<UserInputEvent>();	
	
	protected boolean glInitilised = false; 
	
	protected IScene currentScene; 
	
	public SceneRenderer(View view) {
		this.view = view; 
        this.context = view.getContext();
        
        inputManager = new UserInputManager(view);
        
        this.setCurrentScene(new Scene());
    }
	
	public void onPause(){
		inputManager.onPause();	
	}
	
	public void onResume(){
		startTime = System.nanoTime();
		inputManager.onResume();		
	}
	
	public void setFramerate(long fr){
		this.frameRate = fr; 
	}		
	
	public void setDelegate(ISceneRendererDelegate delegate){
		this.delegate = delegate; 
	}
	
	protected void throttleRenderLoop(){
		long endTime = System.currentTimeMillis();
	    long dt = endTime - lastRenderTimestamp;
	    if (dt < this.frameRate){
	    	try{
	    		Thread.sleep(this.frameRate - dt);
	    	} catch( InterruptedException e){}
	    }
	    lastRenderTimestamp = System.currentTimeMillis();
	}
	
	public boolean isGLInitilised(){
		return glInitilised; 
	}
	
	public void setCurrentScene(IScene scene){
		if( this.currentScene == scene ){
			return; 
		}
		
		if( this.currentScene != null ){
			this.currentScene.setSceneRenderer(null);
			
		}
		
		this.currentScene = scene;
		
		if( this.currentScene != null ){
			this.currentScene.setSceneRenderer(this);		
		}
	}
	
	public IScene getCurrentScene(){
		return this.currentScene; 
	}
	
	/* (non-Javadoc)
	 * @see android.opengl.GLSurfaceView.Renderer#onDrawFrame(javax.microedition.khronos.opengles.GL10)
	 */
	@Override
	public void onDrawFrame(GL10 gl) {
		throttleRenderLoop(); 
		
		// get input 
		inputManager.getInputEvents(inputEvents);
		
		update(); 		
		
		render(); 
	}

	/* (non-Javadoc)
	 * @see android.opengl.GLSurfaceView.Renderer#onSurfaceChanged(javax.microedition.khronos.opengles.GL10, int, int)
	 */
	@Override
	public void onSurfaceChanged(GL10 gl, int width, int height) {
		viewWidth = width; 
		viewHeight = height; 
		GLES20.glViewport(0, 0, viewWidth, viewHeight);
		startTime = System.nanoTime();
		
		if( currentScene != null ){
			currentScene.onSceneRendererSurfaceChanged(this, viewWidth, viewHeight);
		}				
		
		if( delegate != null ){
			delegate.onSurfaceChanged(this);
		}
		
		if( delegate != null ){
			delegate.onLoadSceneEntities(this);
		}
	}

	/* (non-Javadoc)
	 * @see android.opengl.GLSurfaceView.Renderer#onSurfaceCreated(javax.microedition.khronos.opengles.GL10, javax.microedition.khronos.egl.EGLConfig)
	 */
	@Override
	public void onSurfaceCreated(GL10 gl, EGLConfig config) {
		flush(); 
		
		initGL(); 				
		
		glInitilised = true; 
		
		if( delegate != null ){
        	delegate.onSurfaceCreated(this); 
        }				
	}
	
	protected void initGL(){
		//glClearColor(147.0f/255.0f, 204.0f/255.0f, 234.0f/255.0f, 1.0f);
		glClearColor(255.0f/255.0f, 255.0f/255.0f, 255.0f/255.0f, 1.0f);
		
		depthTestEnabled = true; 
		
		if( depthTestEnabled ){
			GLES20.glEnable(GLES20.GL_DEPTH_TEST);
			GLES20.glClearDepthf(1.0f);
			GLES20.glDepthFunc( GLES20.GL_LEQUAL );  // Passes if the incoming depth value is less than or equal to the stored depth value.
			GLES20.glDepthMask( true ); // enable writing into the depth buffer
		}				
		
		
		//if( shadowsEnabled ){
		//	GLES20.glDisable(GLES20.GL_DITHER);	
		//}
	}			
	
	protected void update(){			   
		float elapsedtime = (System.nanoTime()-startTime) / 1000000000.0f;
        startTime = System.nanoTime();                              
        
        if( currentScene != null ){
        	currentScene.onUpdate(elapsedtime);
        }
        
        if( delegate != null ){
        	delegate.onPostUpdate(this, elapsedtime);
        }
	}		 
	
	protected void render(){
		if( !glInitilised || viewWidth == 0 || viewHeight == 0 ){
			return; 
		}
		
		if( currentScene != null ){
			currentScene.onRender(); 
		}
		
		if( delegate != null ){
			delegate.onPostRender(this);
		}
	} 
	
	protected Bitmap takeScreenshot(){
		
		int width = viewWidth;
		int height = viewHeight; 
		
		int b[] = new int[width * height];
		int bt[] = new int[width * height];
		IntBuffer ib = IntBuffer.wrap(b);
		ib.position(0);
		GLES20.glReadPixels(0, 0, width, height, GLES20.GL_RGBA, GLES20.GL_UNSIGNED_BYTE, ib);

		for (int i = 0, k = 0; i < height; i++, k++) {
			for (int j = 0; j < width; j++) {
				int pix = b[i * width + j];
				int pb = (pix >> 16) & 0xff;
				int pr = (pix << 16) & 0x00ff0000;
				int pix1 = (pix & 0xff00ff00) | pr | pb;
				bt[(height - k - 1) * width + j] = pix1;
			}
		}

		Bitmap bitmap = Bitmap.createBitmap(bt, width, height, Bitmap.Config.ARGB_8888);
		
//		try {
//			
//			FileOutputStream fos = new FileOutputStream("/sdcard/screeshot.png");
//			bitmap.compress(Bitmap.CompressFormat.PNG, 100, fos);
//			fos.flush();
//			fos.close();
//		} catch (Exception e) {
//			Log.e("SceneRenderer", e.toString(), e);
//		}				
		
		return bitmap;
	}	
	
	public Context getContext(){
		return context; 
	}
	
	public boolean hasUserInputEvents(){
		return inputEvents.size() > 0; 
	}
	
	public Iterator<UserInputEvent> getUserInputEventsIterator(){
		return inputEvents.iterator();
	}		
	
	public void flush(){
		if( currentScene != null ){
			currentScene.flush(); 
		}
	}

}
