/**
 * ContentLoader.java (ColourExplosion)
 * Created by Joshua Newnham on Apr 9, 2013.
 * Copyright (c) 2013 We Make Play All rights reserved.
 */
package com.wmp.nofrills3d.utils;

import static android.opengl.GLES20.GL_TEXTURE_2D;
import static android.opengl.GLES20.GL_TEXTURE_MAG_FILTER;
import static android.opengl.GLES20.GL_TEXTURE_MIN_FILTER;
import static android.opengl.GLES20.glBindTexture;
import static android.opengl.GLES20.glDeleteTextures;
import static android.opengl.GLES20.glGenTextures;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.IntBuffer;

import android.content.Context;
import android.content.res.Resources;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Matrix;
import android.opengl.GLES20;
import android.opengl.GLUtils;

import com.wmp.nofrills3d.entities.loaders.ModelLoader;
import com.wmp.nofrills3d.entities.loaders.collada.COLLADAModelLoader;
import com.wmp.nofrills3d.entities.loaders.md5.MD5ModelLoader;
import com.wmp.nofrills3d.entities.loaders.obj.ObjModelLoader;
import com.wmp.nofrills3d.material.BufferedTexture;
import com.wmp.nofrills3d.material.Texture;

public class ContentLoader {

	public static final String TAG = "ContentLoader";		
	
	public enum ModelTypeEnum{
		obj, 
		dae, 
		md5
	}
	
	public static ModelLoader createModelLoader(Context context, ModelTypeEnum type){
		if( type == ModelTypeEnum.obj ){
			return new ObjModelLoader(context);
		} 
		else if( type == ModelTypeEnum.dae ){
			return new COLLADAModelLoader(context);
		} 
		else if( type == ModelTypeEnum.md5 ){
			return new MD5ModelLoader(context);
		}
		
		return null; 
	}
	
	public static String loadTextFileFromResource(Context context,
			int resourceId) {
		StringBuilder body = new StringBuilder();

		try {
			InputStream inputStream = context.getResources().openRawResource(
					resourceId);
			InputStreamReader inputStreamReader = new InputStreamReader(
					inputStream);
			BufferedReader bufferedReader = new BufferedReader(
					inputStreamReader);

			String nextLine;

			while ((nextLine = bufferedReader.readLine()) != null) {
				body.append(nextLine);
				body.append('\n');
			}
		} catch (IOException e) {
			throw new RuntimeException(
					"Could not open resource: " + resourceId, e);
		} catch (Resources.NotFoundException nfe) {
			throw new RuntimeException("Resource not found: " + resourceId, nfe);
		}

		return body.toString();
	}
	
	public static Texture loadTextureFromResource(Context context, int resourceId){
		return ContentLoader.loadTextureFromResource(context, resourceId, 
				GLES20.GL_LINEAR, GLES20.GL_LINEAR, 
				GLES20.GL_CLAMP_TO_EDGE, GLES20.GL_CLAMP_TO_EDGE );
	}
	
	public static Texture loadTextureFromResource(Context context, int resourceId, 
			int minFilter, int magFilter, int wrapS, int wrapT ) {
        final int[] textureObjectIds = new int[1];
        int width, height; 
        glGenTextures(1, textureObjectIds, 0);

        if (textureObjectIds[0] == 0) {
        	Logger.LogError(TAG, "Could not generate a new OpenGL texture object.");
            return null;
        } 
        
        final BitmapFactory.Options options = new BitmapFactory.Options();
        options.inScaled = false;

        // Read in the resource
        final Bitmap bitmap = BitmapFactory.decodeResource(context.getResources(), resourceId, options);

        if (bitmap == null) {
        	Logger.LogError(TAG, "Resource ID " + resourceId + " could not be decoded.");
            glDeleteTextures(1, textureObjectIds, 0);
            return null;
        } 
        
        width = bitmap.getWidth(); 
        height = bitmap.getHeight(); 
        
        Matrix flip = new Matrix();
        flip.postScale(1f, -1f);
        Bitmap bitmapFlipped =  Bitmap.createBitmap(bitmap, 0, 0, bitmap.getWidth(), bitmap.getHeight(), flip, true);
        bitmap.recycle();
        
        // Bind to the texture in OpenGL
        glBindTexture(GL_TEXTURE_2D, textureObjectIds[0]);

        // Set filtering: a default must be set, or the texture will be black.
        GLES20.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter);
        GLES20.glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magFilter); 
        GLES20.glTexParameteri(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_WRAP_S, wrapS);
		GLES20.glTexParameteri(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_WRAP_T, wrapT);
        		
        // Load the bitmap into the bound texture.
		GLUtils.texImage2D(GL_TEXTURE_2D, 0, bitmapFlipped, 0);

        // Note: Following code may cause an error to be reported in the
        // ADB log as follows: E/IMGSRV(20095): :0: HardwareMipGen:
        // Failed to generate texture mipmap levels (error=3)
        // No OpenGL error will be encountered (glGetError() will return
        // 0). If this happens, just squash the source image to be
        // square. It will look the same because of texture coordinates,
        // and mipmap generation will work.

        //glGenerateMipmap(GL_TEXTURE_2D);

        // Recycle the bitmap, since its data has been loaded into OpenGL.
		bitmapFlipped.recycle();

        // Unbind from the texture.
        glBindTexture(GL_TEXTURE_2D, 0);

        return new Texture(textureObjectIds[0], width, height);
    }
	
	public static BufferedTexture createTexture(int width, int height){
		return ContentLoader.createTexture(width, height, 
				GLES20.GL_LINEAR, GLES20.GL_LINEAR, 
				GLES20.GL_CLAMP_TO_EDGE, GLES20.GL_CLAMP_TO_EDGE);
	}
	
	public static BufferedTexture createTexture(int width, int height, 
			int minFilter, int magFilter, int wrapS, int wrapT ) {
		
		final int FLOAT_SIZE_BYTES = 4;
		
		int[] renderTex = new int[1];
		GLES20.glGenTextures(1, renderTex, 0);
		
		int textureId = renderTex[0]; 
		
		GLES20.glBindTexture(GLES20.GL_TEXTURE_2D, textureId);
		
		GLES20.glTexParameteri(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_MIN_FILTER, minFilter);
		GLES20.glTexParameteri(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_MAG_FILTER, magFilter);		
		
		GLES20.glTexParameteri(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_WRAP_S, wrapS);
		GLES20.glTexParameteri(GLES20.GL_TEXTURE_2D, GLES20.GL_TEXTURE_WRAP_T, wrapT);
		
		//int[] buf = new int[_width * _height];
		int bufLen = width * height;	
		IntBuffer _texBuffer = ByteBuffer.allocateDirect(bufLen * FLOAT_SIZE_BYTES).order(ByteOrder.nativeOrder()).asIntBuffer();
		
		GLES20.glTexImage2D(GLES20.GL_TEXTURE_2D, 0, GLES20.GL_RGBA, width, height, 0, GLES20.GL_RGBA, GLES20.GL_UNSIGNED_BYTE, _texBuffer);		
		
		BufferedTexture texture = new BufferedTexture(textureId, width, height);
		texture._texBuffer = _texBuffer;
		
		GLES20.glBindTexture(GLES20.GL_TEXTURE_2D, 0);
		
		return texture; 
		
	}
}
